package sos.scheduler.job;

import java.io.File;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import sos.net.SOSMail;
import sos.spooler.Order;
import sos.spooler.Variable_set;
import sos.util.SOSDate;
import sos.util.SOSDocumentFactoryTextProcessor;
import sos.util.SOSPlainTextProcessor;
import sos.util.SOSTextProcessor;

/**
 * process mails
 * 
 * job parameters:
 * table_mails
 * table_mail_attachments
 * table_settings
 * application_mail
 * section_mail
 * application_mail_templates
 * section_mail_templates
 * mail_to
 * queue_directory
 * log_directory
 * log_only
 * 
 * order parameters:
 * id order identification
 * 
 * @author andreas.pueschel@sos-berlin.com
 * @since 1.0 2006-01-03 
 */
public class JobSchedulerSendMailJob extends JobSchedulerMailJob {

    protected Vector mailOrders                     = null;
    
    protected Iterator mailOrderIterator            = null;

    protected String encoding                       = "7bit";
    protected String charset                        = "iso-8859-1";

    protected String attachmentEncoding             = "Base64";
    protected String attachmentCharset              = "iso-8859-1";
    
    protected SOSTextProcessor mailTextProcessor                                = null;
    
    protected SOSPlainTextProcessor mailPlainTextProcessor                      = null;
    
    protected SOSDocumentFactoryTextProcessor mailDocumentFactoryTextProcessor  = null;

    protected boolean hasDatabase                   = true;

    private String lastLanguage                      = "de";
    
    
    /**
     * initialization 
     */
    public boolean spooler_init() {

        if (!super.spooler_init()) return false;
        
        try { // initialize text processors with templates from settings
            this.mailPlainTextProcessor = new SOSPlainTextProcessor(this.sosMailSettings, this.getLogger());
            this.mailPlainTextProcessor.setHasLocalizedTemplates(this.hasLocalizedTemplates());
            this.mailPlainTextProcessor.getTemplates(this.getSectionMailTemplates(), this.getApplicationMailTemplates());
            this.mailPlainTextProcessor.getScripts(this.getSectionMailScripts(), this.getApplicationMailScripts());
            
            this.mailDocumentFactoryTextProcessor = new SOSDocumentFactoryTextProcessor(this.sosMailSettings, this.getLogger());
            this.mailDocumentFactoryTextProcessor.setHasLocalizedTemplates(this.hasLocalizedTemplates());
            this.mailDocumentFactoryTextProcessor.getTemplates(this.getSectionMailTemplatesFactory(), this.getApplicationMailTemplatesFactory());
            this.mailDocumentFactoryTextProcessor.getScripts(this.getSectionMailScripts(), this.getApplicationMailScripts());
            
            return true;
            
        } catch (Exception e){
            spooler_log.warn("failed to initialize job: " + e.getMessage());
            return false;
        }
    }
    
    
    /**
     * select mails to process for classic job mode 
     */
    public boolean spooler_open() {
        
        try {
            // select mails for classic job mode
            if (spooler_task.job().order_queue() == null) {
                this.mailOrders = this.getConnection().getArrayAsVector("SELECT \"ID\", \"MAILING_ID\", \"JOB_ID\", \"MESSAGE_ID\", \"CLIENT_IDENTIFIER\", \"TOPIC\", \"TOPIC_IDENTIFIER\", \"MAIL_FROM\", \"FROM_NAME\","
                    + " \"MAIL_TO\", \"CC_TO\", \"BCC_TO\", \"REPLY_TO\", \"PRIORITY\", \"SUBJECT\", \"SUBJECT_TEMPLATE\", \"SUBJECT_TEMPLATE_TYPE\","
                    + " \"BODY\", \"BODY_TEMPLATE\", \"BODY_TEMPLATE_TYPE\", \"REPLACEMENTS\", \"LANGUAGE\", \"CHARSET\", \"ENCODING\", \"CONTENT_TYPE\", \"STATUS\", \"STATUS_TEXT\","
                    + " \"TARGETED\", \"DELIVERED\", \"CREATED\", \"CREATED_BY\", \"MODIFIED\", \"MODIFIED_BY\" FROM " + this.getTableMails() 
                    + " WHERE \"STATUS\"=" + STATE_READY + " AND (\"TARGETED\" IS NULL OR \"TARGETED\"<=%now) ORDER BY \"ID\" ASC");

                this.mailOrderIterator = this.mailOrders.iterator();
                return this.mailOrderIterator.hasNext();
            } else {
                return true;
            }
            
        } catch (Exception e){
            spooler_log.warn("failed to retrieve mail orders: " + e.getMessage());
            return false;
        }
        finally {
            try {
                if (this.getConnection() != null) this.getConnection().rollback();
            } 
            catch (Exception ex) {} // gracefully ignore this error to preserve the original exception
        }
    }
    
    
    /**
     * process single mail 
     */
    public boolean spooler_process() {

        Order order = null;
        Variable_set orderData = null;
        HashMap mailOrder = null;
        int mailOrderId = 0;
        boolean rc = true;
        
        try {
            // classic or order driven
            if (spooler_task.job().order_queue() == null) {
                if (!this.mailOrderIterator.hasNext()) return true;
                mailOrder = (HashMap) this.mailOrderIterator.next();
                rc = this.mailOrderIterator.hasNext();

                if (mailOrder.isEmpty()) 
                    throw new Exception("empty mail order was found.");
            } else {
                order = spooler_task.order();
                orderData = (Variable_set) spooler_task.order().payload();
                
                if ( orderData.var("id") == null || orderData.var("id").toString().length() == 0)
                    throw (new Exception("no order id was given in payload"));
                
                mailOrder = this.getConnection().getSingle("SELECT \"ID\", \"MAILING_ID\", \"JOB_ID\", \"MESSAGE_ID\", \"CLIENT_IDENTIFIER\", \"TOPIC\", \"TOPIC_IDENTIFIER\", \"MAIL_FROM\", \"FROM_NAME\","
                        + " \"MAIL_TO\", \"CC_TO\", \"BCC_TO\", \"REPLY_TO\", \"PRIORITY\", \"SUBJECT\", \"SUBJECT_TEMPLATE\",\"SUBJECT_TEMPLATE_TYPE\","
                        + " \"BODY\", \"BODY_TEMPLATE\", \"BODY_TEMPLATE_TYPE\", \"REPLACEMENTS\", \"LANGUAGE\", \"CHARSET\", \"ENCODING\", \"CONTENT_TYPE\", \"STATUS\", \"STATUS_TEXT\","
                        + " \"TARGETED\", \"DELIVERED\", \"CREATED\", \"CREATED_BY\", \"MODIFIED\", \"MODIFIED_BY\" FROM " + this.getTableMails() 
                        + " WHERE \"ID\"=" + orderData.var("id")); 

                if (mailOrder.isEmpty())  {
                    if (this.getLogger() != null) this.getLogger().info("no order was found for this id: " + orderData.var("id"));
                    return false;
                }

                // setback this order for future target date
                if (mailOrder.get("targeted") != null 
                    && mailOrder.get("targeted").toString().length() > 0) {
                    Date currentDate = SOSDate.getTime();
                    Date targetDate = SOSDate.getTime(mailOrder.get("targeted").toString());
                    if (targetDate.after(currentDate)) {
                        Calendar target = Calendar.getInstance();
                        target.setTime(targetDate);
                        spooler_task.job().set_delay_order_after_setback(1, (target.getTimeInMillis()-System.currentTimeMillis())/1000);
                        spooler_task.order().setback();
                        if (this.getLogger() != null) this.getLogger().info("order is set back for target date: " + mailOrder.get("targeted").toString());
                        return false;
                    }
                }
            }

            
            try { // to check status: allow all status values except STATE_SUCCESS and STATE_CANCEL
                if (mailOrder.get("status") != null 
                        && mailOrder.get("status").toString().length() > 0) {
                    int status = Integer.parseInt(mailOrder.get("status").toString());
                    switch (status) {
                        case STATE_SUCCESS: throw new Exception("mail order has already been successfully processed");
                        case STATE_CANCEL:  throw new Exception("mail order has already been cancelled"); 
                    }
                }
            } catch (Exception ex) {
                if (this.getLogger() != null) this.getLogger().info("mail status [" + mailOrder.get("status").toString() + "] not applicable for order [" + mailOrder.get("id").toString() + "]: " + ex.getMessage());
                // return value for classic and order driven processing
                if (spooler_task.job().order_queue() == null) {
                    return this.mailOrderIterator.hasNext();
                } else {
                    return false;
                }
            }

            
            // create mailer with settings
            sosMail = new SOSMail( this.sosMailSettings );

            if (this.getLogger() != null) sosMail.setSOSLogger(this.getLogger());
            
            // job scheduler sets mail host if no default was specified by settings
            if (sosMail.getHost() == null || sosMail.getHost().length() == 0)
                sosMail.setHost(spooler_log.mail().smtp());
            
            // job scheduler sets mail queue directory if no default was specified by parameters or settings
            if (sosMail.getQueueDir() == null || sosMail.getQueueDir().length() == 0)
                sosMail.setQueueDir(this.getQueueDirectory());

            // set queue prefix "sos" to enalbe dequeueing by JobSchedulerMailDequeueJob
            sosMail.setQueuePraefix(this.getQueuePrefix());

            
            if (mailOrder.get("id") != null 
                    && mailOrder.get("id").toString().length() > 0) {
                try {
                    mailOrderId = Integer.parseInt(mailOrder.get("id").toString());
                } catch (Exception e) {
                    throw new Exception("illegal value for order id in payload: " + mailOrder.get("id").toString());
                }
            } else {
                throw new Exception("no id was found in this order payload.");
            }


            if (mailOrder.get("mail_queue_directory") != null 
                    && mailOrder.get("mail_queue_directory").toString().length() > 0)
                    sosMail.setQueueDir(mailOrder.get("mail_queue_directory").toString());
            else if (mailOrder.get("queue_directory") != null 
                    && mailOrder.get("queue_directory").toString().length() > 0)
                    sosMail.setQueueDir(mailOrder.get("queue_directory").toString());

            if (mailOrder.get("mail_log_directory") != null 
                    && mailOrder.get("mail_log_directory").toString().length() > 0)
                    this.setLogDirectory(mailOrder.get("mail_log_directory").toString());
            else if (mailOrder.get("log_directory") != null 
                    && mailOrder.get("log_directory").toString().length() > 0)
                    this.setLogDirectory(mailOrder.get("log_directory").toString());
            
            if (mailOrder.get("mail_log_only") != null 
                    && mailOrder.get("mail_log_only").toString().length() > 0)
                    this.setLogOnly((mailOrder.get("mail_log_only").toString().equals("1")
                            || mailOrder.get("mail_log_only").toString().equalsIgnoreCase("true")
                            || mailOrder.get("mail_log_only").toString().equalsIgnoreCase("yes")) ? true : false);
            
            if (mailOrder.get("log_only") != null 
                    && mailOrder.get("log_only").toString().length() > 0)
                    this.setLogOnly((mailOrder.get("log_only").toString().equals("1")
                            || mailOrder.get("log_only").toString().equalsIgnoreCase("true")
                            || mailOrder.get("log_only").toString().equalsIgnoreCase("yes")) ? true : false);
            
            
            if (mailOrder.get("mail_from") != null 
                    && mailOrder.get("mail_from").toString().length() > 0)
                sosMail.setFrom(mailOrder.get("mail_from").toString());

            if (mailOrder.get("from_name") != null 
                    && mailOrder.get("from_name").toString().length() > 0)
                sosMail.setFromName(mailOrder.get("from_name").toString());

            if (mailOrder.get("reply_to") != null 
                    && mailOrder.get("reply_to").toString().length() > 0)
                sosMail.setReplyTo(mailOrder.get("reply_to").toString());

            if (mailOrder.get("priority") != null 
                    && mailOrder.get("priority").toString().length() > 0) {
                switch(Integer.parseInt(mailOrder.get("priority").toString())) {
                    case 1: sosMail.setPriorityHigh(); break;
                    case 2: sosMail.setPriorityLow(); break;
                    case 3: sosMail.setPriorityLowest(); break;
                    default: sosMail.setPriorityNormal(); break;
                }
            }

            if (mailOrder.get("charset") != null 
                    && mailOrder.get("charset").toString().length() > 0)
                sosMail.setCharset(mailOrder.get("charset").toString());
                
            if (mailOrder.get("encoding") != null 
                    && mailOrder.get("encoding").toString().length() > 0)
                sosMail.setEncoding(mailOrder.get("encoding").toString());
                
            if (mailOrder.get("content_type") != null 
                    && mailOrder.get("content_type").toString().length() > 0)
                sosMail.setContentType(mailOrder.get("content_type").toString());

            HashMap headers = new HashMap();
            if (mailOrder.get("headers") != null 
                    && mailOrder.get("headers").toString().length() > 0) {
                String[] headerList = mailOrder.get("headers").toString().split("\\,");
                for(int i=0; i<headerList.length; i++) {
                    String[] headerEntry = headerList[i].split("\\:");
                    if (headerEntry.length > 0) {
                        headers.put(headerEntry[0], headerEntry[1]);
                    } else {
                        throw new Exception("illegal header syntax [" + headerEntry[0] + "], use: name:value,name:value");
                    }
                }
            }

            
            if (mailOrder.get("mail_to") != null
                    && mailOrder.get("mail_to").toString().length() > 0) {
                sosMail.addRecipient(mailOrder.get("mail_to").toString());
            }
                    
            if (mailOrder.get("cc_to") != null
                    && mailOrder.get("cc_to").toString().length() > 0) {
                sosMail.addCC(mailOrder.get("cc_to").toString());
            }
                
            if (mailOrder.get("bcc_to") != null 
                    && mailOrder.get("bcc_to").toString().length() > 0) {
                sosMail.addBCC(mailOrder.get("bcc_to").toString());
            }

            if (mailOrder.get("language") != null 
                    && mailOrder.get("language").toString().length() > 0)
                sosMail.setLanguage(mailOrder.get("language").toString());

            HashMap replacements = new HashMap();
            if (mailOrder.get("replacements") != null 
                    && mailOrder.get("replacements").toString().length() > 0) {
                String[] replacementList = mailOrder.get("replacements").toString().split("\\|");
                for(int i=0; i<replacementList.length; i++) {
                    String[] replacementEntry = replacementList[i].split("\\^");
                    replacements.put(replacementEntry[0], replacementEntry[1]);
                }
            }

            
            try {
                sosMail.setAttachmentEncoding(this.attachmentEncoding);
                sosMail.setAttachmentCharset(this.attachmentCharset);
                
                Vector attachmentOrders = this.getConnection().getArrayAsVector("SELECT \"FILENAME\", \"ENCODING\", \"CHARSET\", \"CONTENT_TYPE\" FROM "
                        + this.getTableMailAttachments() + " WHERE \"ID\"=" + mailOrderId);
                Iterator attachmentIterator = attachmentOrders.iterator();
                while (attachmentIterator.hasNext()) {
                    HashMap attachment = (HashMap)attachmentIterator.next();
                    if (attachment != null) {
                        String attParams = "";
                        String attConc = "";
                        if (attachment.get("content_type") != null && attachment.get("content_type").toString().length() > 0) {
                            attParams += attConc + "content-type=" + attachment.get("content_type").toString();
                            attConc = ",";
                        }
                        if (attachment.get("encoding") != null && attachment.get("encoding").toString().length() > 0) {
                            attParams += attConc + "encoding=" + attachment.get("encoding").toString();
                            attConc = ",";
                        }
                        if (attachment.get("charset") != null && attachment.get("charset").toString().length() > 0) {
                            attParams += attConc + "charset=" + attachment.get("charset").toString();
                            attConc = ",";
                        }

                        sosMail.addAttachment(attachment.get("filename").toString(), attParams);
                    }
                }
            } catch (Exception ex) {
                throw new Exception("error occurred processing attachment: " + ex.getMessage());
            }

            
            try { // to call your local implementation of prepare()
                sosMail = this.prepare(sosMail, mailOrder);
            } catch (Exception ex) {
                throw new Exception("error in prepare(): " + ex.getMessage());
            }

            
            try { // to call your local implementation of prepareReplacements()
                replacements = this.prepareReplacements(sosMail, mailOrder, replacements);
            } catch (Exception ex) {
                throw new Exception("error in prepareReplacements(): " + ex.getMessage());
            }

            // set subject if this has not been done by prepare()
            if (sosMail.getSubject() == null || sosMail.getSubject().trim().length() == 0) {
                if (mailOrder.get("subject") != null 
                        && mailOrder.get("subject").toString().length() > 0) {
                    sosMail.setSubject(mailOrder.get("subject").toString());
                } else if (mailOrder.get("subject_template") != null 
                        && mailOrder.get("subject_template").toString().length() > 0) {
                    if (mailOrder.get("subject_template_type") != null && mailOrder.get("subject_template_type").toString().equals("1")) {
                        // use Document Factory Text Processor
                        if (!sosMail.getLanguage().equals(this.getLastLanguage())) {
                            this.setLastLanguage(sosMail.getLanguage());
                            this.mailDocumentFactoryTextProcessor.setLanguage(sosMail.getLanguage());
                            this.mailDocumentFactoryTextProcessor.initLanguage();
                            this.mailDocumentFactoryTextProcessor.setForceReload(true);
                            this.mailDocumentFactoryTextProcessor.getTemplates(this.getSectionMailTemplatesFactory(), this.getApplicationMailTemplatesFactory());
                        }
                        sosMail.setSubject(this.mailDocumentFactoryTextProcessor.process(this.mailDocumentFactoryTextProcessor.getTemplate(mailOrder.get("subject_template").toString()), replacements));
                    } else {
                        // use Plain Text Processor
                        if (!sosMail.getLanguage().equals(this.getLastLanguage())) {
                            this.setLastLanguage(sosMail.getLanguage());
                            this.mailPlainTextProcessor.setLanguage(sosMail.getLanguage());
                            this.mailPlainTextProcessor.initLanguage();
                            this.mailPlainTextProcessor.setForceReload(true);
                            this.mailPlainTextProcessor.getTemplates(this.getSectionMailTemplates(), this.getApplicationMailTemplates());
                        }
                        sosMail.setSubject(this.mailPlainTextProcessor.process(this.mailPlainTextProcessor.getTemplate(mailOrder.get("subject_template").toString()), replacements, (sosMail.getContentType().equals("text/html") ? true : false)));
                    }
                }
            }
                
            // set body if this has not been done by prepare()
            if (sosMail.getBody() == null || sosMail.getBody().trim().length() == 0) {
                if (mailOrder.get("body") != null 
                        && mailOrder.get("body").toString().length() > 0) {
                    sosMail.setBody(mailOrder.get("body").toString());
                } else if (mailOrder.get("body_template") != null 
                        && mailOrder.get("body_template").toString().length() > 0) {
                    if (mailOrder.get("body_template_type") != null && mailOrder.get("body_template_type").toString().equals("1")) { 
                        // use Document Factory Text Processor
                        if (!sosMail.getLanguage().equals(this.getLastLanguage())) {
                            this.setLastLanguage(sosMail.getLanguage());
                            this.mailDocumentFactoryTextProcessor.setLanguage(sosMail.getLanguage());
                            this.mailDocumentFactoryTextProcessor.initLanguage();
                            this.mailDocumentFactoryTextProcessor.setForceReload(true);
                            this.mailDocumentFactoryTextProcessor.getTemplates(this.getSectionMailTemplatesFactory(), this.getApplicationMailTemplatesFactory());
                        }
                        sosMail.setBody(this.mailDocumentFactoryTextProcessor.process(this.mailDocumentFactoryTextProcessor.getTemplate(mailOrder.get("body_template").toString()), replacements));
                    } else {
                        // use Plain Text Processor
                        if (!sosMail.getLanguage().equals(this.getLastLanguage())) {
                            this.setLastLanguage(sosMail.getLanguage());
                            this.mailPlainTextProcessor.setLanguage(sosMail.getLanguage());
                            this.mailPlainTextProcessor.initLanguage();
                            this.mailPlainTextProcessor.setForceReload(true);
                            this.mailPlainTextProcessor.getTemplates(this.getSectionMailTemplates(), this.getApplicationMailTemplates());
                        }
                        sosMail.setBody(this.mailPlainTextProcessor.process(this.mailPlainTextProcessor.getTemplate(mailOrder.get("body_template").toString()), replacements, (sosMail.getContentType().equals("text/html") ? true : false)));
                    }
                }
            }


            // check required fields
            if (sosMail.getSubject() == null || sosMail.getSubject().trim().length() == 0) {
                throw new Exception("empty subject not allowed");
            }

            if (sosMail.getBody() == null || sosMail.getBody().trim().length() == 0) {
                throw new Exception("empty body not allowed");
            }


            try {
                File mailFile  = null;
                String message = "";

                // clear recipients for test purposes
                if (this.getMailTo() != null && this.getMailTo().length() > 0) {
                    if (this.getLogger() != null) this.getLogger().info("email recipients [" + sosMail.getRecipientsAsString() + "] are replaced by settings: " + this.getMailTo());
                    sosMail.clearRecipients();
                    sosMail.addRecipient(this.getMailTo());
                }

                if (this.getLogDirectory() != null && this.getLogDirectory().length() > 0) {
                    mailFile = this.getMailFile(this.getLogDirectory());
                    sosMail.dumpMessageToFile(mailFile, true);
                }
                
                if (this.isLogOnly()) {
                    message = "mail was NOT sent but stored to file: " + mailFile.getAbsolutePath();
                    if (this.getLogger() != null) this.getLogger().info(message);
                    message = (message.length() > 250 ? message.substring(message.length()-250) : message);
                    message = message.replaceAll("'", "''");
                } else if (!sosMail.send()) {
                    message = "mail was NOT sent but stored for later dequeueing, reason was: " + sosMail.getLastError();
                    if (this.getLogger() != null) this.getLogger().info(message);
                    message = (message.length() > 250 ? message.substring(message.length()-250) : message);
                    message = message.replaceAll("'", "''");
                }

                if (this.getLogger() != null) this.getLogger().info("mail was " + (this.isLogOnly() ? "processed" : "sent") + " for order " + mailOrderId + " to: " + sosMail.getRecipientsAsString());
                if (this.getConnection() != null && mailOrderId > 0) {
                    this.getConnection().executeUpdate("UPDATE " + this.getTableMails() + " SET \"JOB_ID\"=" + spooler_task.id() + ", \"STATUS\"=" + STATE_SUCCESS + ", \"STATUS_TEXT\"='" + message + "'"
                        + ", \"MESSAGE_ID\"='" + sosMail.getMessage().getMessageID() + "'"
                        + ", \"MODIFIED\"=%now, \"MODIFIED_BY\"='" + spooler_task.job().name() + "', \"DELIVERED\"=%now"
                        + " WHERE \"ID\"=" + mailOrderId);
                    this.getConnection().commit();
                }

                if (this.getLogger() != null) this.getLogger().debug3("mail was sent with headers: " + sosMail.dumpHeaders());
            } catch (Exception ex) {
                throw new Exception("mail was NOT sent for order " + mailOrderId + ": " + ex.getMessage());
            }


            // return value for classic and order driven processing
            if (spooler_task.job().order_queue() == null) {
                return this.mailOrderIterator.hasNext();
            } else {
                return rc;
            }
        }    
        catch (Exception e) {
            if (spooler_task.job().order_queue() != null) {
                spooler_log.warn("error occurred processing mail [" + ((order != null) ? "job chain: " + order.job_chain().name() + ", order: " + order.id() : "(none)") + "]: " + e.getMessage());
            } else { 
                spooler_log.warn("error occurred processing mail: " + e.getMessage());
            }

            try { // to update database if mail order has been identified
                if (this.getConnection() != null && mailOrderId > 0) {
                    String message = "";
                    if (e.getMessage() != null) {
                        message = (e.getMessage().length() > 250 ? e.getMessage().substring(e.getMessage().length()-250) : e.getMessage());
                    } else {
                        message = e.toString();
                    }
                    message = message.replaceAll("'", "''");
                    this.getConnection().executeUpdate("UPDATE " + this.getTableMails() + " SET \"JOB_ID\"=" + spooler_task.id() + ", \"STATUS\"=" + STATE_ERROR + ", \"STATUS_TEXT\"='" + message + "',"
                            + "\"MODIFIED\"=%now, \"MODIFIED_BY\"='" + spooler_task.job().name() + "', \"DELIVERED\"=NULL"
                            + " WHERE \"ID\"=" + mailOrderId);
                    this.getConnection().commit();
                }
            } catch (Exception ex) {
                spooler_log.warn("error occurred updating mail database: " + ex.getMessage());
            }
            
            // restart this order driven task in case of errors to make the task log available by mail
            if (spooler_task.job().order_queue() != null) {            
                spooler_task.end();
                return false;
            } else {
                return this.mailOrderIterator.hasNext();
            }
        } 
        finally {
            try {
                if (this.getConnection() != null) this.getConnection().rollback();
            } 
            catch (Exception ex) {} // gracefully ignore this error to preserve the original exception
        }
	}


    /**
     * overwrite this method to prepare the message object
     * @param sosMail
     * @return SOSMail
     * @throws Exception
     */
    public SOSMail prepare(SOSMail sosMail, HashMap mailOrder) throws Exception {
        
        return sosMail;
    }

    /**
     * overwrite this method to prepare the message object
     * @param sosMail
     * @return SOSMail
     * @throws Exception
     */
    public HashMap prepareReplacements(SOSMail sosMail, HashMap mailOrder, HashMap replacements) throws Exception {
        
        return replacements;
    }


    /**
     * @return Returns the lastLanguage.
     */
    public String getLastLanguage() {
        return lastLanguage;
    }


    /**
     * @param lastLanguage The lastLanguage to set.
     */
    public void setLastLanguage(String lastLanguage) {
        this.lastLanguage = lastLanguage;
    }


    /**
     * @return Returns the hasDatabase.
     */
    public boolean hasDatabase() {
        return hasDatabase;
    }



    /**
     * @param hasDatabase The hasDatabase to set.
     */
    public void setHasDatabase(boolean hasDatabase) {
        this.hasDatabase = hasDatabase;
    }
}