package sos.settings.converter;


import java.util.*;

import sos.util.*;
import sos.settings.*;
import java.io.*;

/**
 * <p>Title: sos.settings.converter.SOSSettingsXML2ConnectionConverter.java </p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 09.07.2004</p>
 * <p>Company: SOS GmbH</p>
 * @author Mrvet ksz
 * @resource 
 */
public class SOSSettingsXML2ConnectionConverter extends SOSSettingsConverter {

    /**
     * Attribut: SOSStandardLogger Objekt.
     * @see sos.util.SOSStandardLogger 
     * */
    private SOSStandardLogger sosLogger = null;
    
    /** Attribut: Name der Konfigurationsdatei*/
    private String configFile = "";
        
    /** Attribut: Alle Eintrge der Sektion*/
    private ArrayList sections = null;
    
    /** Attribut: BufferWriter */
    private BufferedWriter output = null;
       
    /** Attribut: SOSString Objekt */
    private SOSString sosString = null;
    
    /** Attribut: SOSProfileSettings Object*/
    private SOSXMLSettings settings = null;

    /** Attribut: encoding; default ist iso-8859-1 */
    private String encoding = "iso-8859-1";
    
    /** Attribut: namespace*/
    private String namespace = "";
    
    /** Attribut: Name der Tabelle*/
    private String tablename = "SETTINGS"; 
    
    /**
     * Konstruktor
     * @exception Exception
     */
    public SOSSettingsXML2ConnectionConverter()throws Exception {
        super();        
    }

    /**
     * @param sosLogger_
     * @throws Exception
     */
    public SOSSettingsXML2ConnectionConverter(SOSStandardLogger sosLogger_)
            throws Exception {
        super(sosLogger_);
    }
    
    /**
     * Konstruktor.
     * 
     * @param SOSStandardLogger
     * @see sos.util.SOSStandardLogger
     * @throws Exception
     */
    public SOSSettingsXML2ConnectionConverter(String configFile_, SOSStandardLogger sosLogger_)
    		throws Exception {
        super(configFile_, sosLogger_);
        
    }
    
    /**
     * Ldt die XML-Konfigurationsdatei. <br>
     * Diese Methode liefert true, wenn die Konfigurationsdatei erfolgreich 
     * geladen wurde. 
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean loadSettings() throws Exception {
        boolean retVal  =true;
        try {
            settings = new SOSXMLSettings(configFile, sosLogger);  
	        sections = settings.getListOfXMLFields();            
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    /**
     * Schreibt die Konfigurationsdatei in eine neue Konfigurationsdatei. 
     * -> von XML zu INI Datei.
     * Wenn die INI-Konfigutaionsdatei erfolgreich geschrieben wurde, dann 
     * liefert die Methode true.
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean writeSettingFile() throws Exception {
        boolean retVal  =true;
        String sectionname = "";
        Properties entrys = null;
        try {
            sosLogger.info("Creating XML-Connection-Settings-File from config File " + configFile);  
            createWriterObject();           
            writeXMLConnectionFile();
            output.close();
            sosLogger.info("XML-Settings-File is successfully created:" + configFile.replaceAll("\\.xml", "_connection\\.xml"));                        
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    /**
     * Schreibt die XML-ConnectionSettings-Datei.
     * Diese Datei kann dann ber die Weboberflche in die Datenbank importiert werden.
     * 
     * 
     * @throws Exception
     */
    private void writeXMLConnectionFile() throws Exception {                    
            String sectionname = "";
            HashMap entrys = null;
            boolean buildAppName = false;
            String application = "";
            try {
                
                if (output == null) {
                    throw new Exception("..Bufferwriter Object ist null.");                
                }
                
                sosString = new SOSString();
                
                //Wenn die Applikationsname nicht bergeben wurde, dann wird diese hier 
                //in der Form application_sectionname gebildet.
                entrys = (HashMap)sections.get(0);
                if (entrys != null && entrys.size() > 0) {
                    application = sosString.parseToString(entrys,("application"));
                }
                
    			output.write("<?xml version=\"1.0\" encoding=\"" + getEncoding() +"\"?>");    			
    			output.write("<sos_export " + namespace + " application=\"" + application + "\" created=\"" + sos.util.SOSDate.getCurrentDateAsString() +"\">");

    			output.write("<sos_export_package id=\""+ this.getTablename()+"\">");
    			output.write("<sos_export_meta>");
    				output.write("<table name=\""+ getTablename()+"\" />");
    				
    				output.write("<key_fields>");
    					output.write("<field name=\"application\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
    					output.write("<field name=\"section\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
    					output.write("<field name=\"name\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
    				output.write("</key_fields>");
    				
    				output.write("<fields>");
    					output.write("<field name=\"application\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
    					output.write("<field name=\"section\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
    					output.write("<field name=\"name\" type=\"VARCHAR\" len=\"250\" scale=\"\" /> ");
    					output.write("<field name=\"value\" type=\"VARCHAR\" len=\"1000\" scale=\"\" /> ");
    					output.write("<field name=\"default_value\" type=\"VARCHAR\" len=\"1000\" scale=\"\" /> ");
    					output.write("<field name=\"long_value\" type=\"BLOB\" len=\"4000\" scale=\"\" /> ");
    					output.write("<field name=\"title\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
    					output.write("<field name=\"input_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"input_size\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"display_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"display_size\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"forced\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"entry_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
    					output.write("<field name=\"created\" type=\"DATE\" len=\"7\" scale=\"\" /> ");
    					output.write("<field name=\"created_by\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
    					output.write("<field name=\"modified\" type=\"DATE\" len=\"7\" scale=\"\" /> ");
    					output.write("<field name=\"modified_by\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
    				output.write("</fields>");
    				
    				output.write("</sos_export_meta>");
    			        		
    			    output.write("<sos_export_data>");
    			    for (int i = 0; i < this.sections.size(); i++) { 			    
    	                entrys = (HashMap)sections.get(i);      			        
    			        
    			            output.write("<sos_export_record name=\"" + tablename +"\">");
    					            output.write("<sos_export_fields>");
    					            output.write("<application><![CDATA[" + sosString.parseToString(entrys.get("application")) +"]]></application>");
    					            output.write("<section><![CDATA[" + sosString.parseToString(entrys.get("section")) +"]]></section>");
    					            output.write("<name><![CDATA[" + sosString.parseToString(entrys.get("name")) +"]]></name>");
    					            output.write("<value><![CDATA[" + sosString.parseToString(entrys.get("value")) +"]]></value>");
    					            output.write("<default_value></default_value>");
    					            output.write("<long_value></long_value>");
    					            output.write("<title>"+sosString.parseToString(entrys.get("section"))+"</title>");
    					            output.write("<input_type>0</input_type>");
    					            output.write("<input_size>0</input_size>");
    					            output.write("<display_type>0</display_type>");
    					            output.write("<display_size>10</display_size>");
    					            output.write("<forced>0</forced>");
    					            output.write("<entry_type>0</entry_type>");
    					            output.write("<created>"+SOSDate.getCurrentDateAsString() + "</created>");
    					            output.write("<created_by></created_by>");
    					            output.write("<modified>"+SOSDate.getCurrentDateAsString() +"</modified>");
    					            output.write("<modified_by></modified_by>");
    				            output.write("</sos_export_fields>");
    			            output.write("</sos_export_record>");
    			        }
    			        	
    			   	output.write("</sos_export_data>");
    			  output.write("</sos_export_package>");		
    			output.write("</sos_export>");			                            
            } catch (Exception e) {
                throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
            }
        }
    
    

    /**
     * Erstellt das BufferedWriter Objekt. 
     * 
     * @throws Exception
     */
    private void createWriterObject() throws Exception {       
        String fileName = "";
        try {
            fileName = this.configFile.replaceAll("\\.xml", "_connection\\.xml");                       
            try {
                output = new BufferedWriter((new FileWriter(fileName)));
            } catch (IOException e) {
                throw new Exception(
                        "..error while creating Bufferwriter in createWriterObject: " + e);
            }                        
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    /**
     * Diese Methode liefert den Wert der Property <i>configFile</i>. 
     * Der Typ der Property <i>configFile</i> ist String.
     * Die configfile ist dioe XML-Konfigurationsdatei, die in INI-Konfigurationsdatei
     * geschrieben werden soll.
     * 
     * @return String.
     */
    public String getConfigFile() {
        return configFile;
    }
    /**
     * Hier wird ein Wert zu den Property <i>configFile</i> gesetzt.
     * Der Typ der Property <i>configFile</i> ist <i>String</i>.
     *
     * @param configFile
     */
    public void setConfigFile(String configFile) {
        this.configFile = configFile;
    }
    
    /**
     * Hier wird ein Wert zu den Property <i>sosLogger</i> gesetzt.
     * Der Typ der Property <i>sosLogger</i> ist <i>SOSStandardLogger</i>.
     *
     * @see sos.util.SOSStandardLogger
     * @param sosLogger
     */
    public void setSosLogger(SOSStandardLogger sosLogger) {
        this.sosLogger = sosLogger;
    }

    /**
     * Diese Methode liefert den Wert der Property <i>tablename</i>. 
     * Der Typ der Property <i>tablename</i> ist String.
     * 
     * @return String.
     */
    public String getTablename() {
        return tablename;
    }
    /**
     * Hier wird ein Wert zu den Property <i>tablename</i> gesetzt.
     * Der Typ der Property <i>tablename</i> ist <i>String</i>.
     *
     * @param tablename
     */
    public void setTablename(String tablename) {
        this.tablename = tablename;
    }
    
    /**
     * Diese Methode liefert den Wert der Property <i>encoding</i>. 
     * Der Typ der Property <i>encoding</i> ist String.
     * 
     * @return String.
     */
    public String getEncoding() {
        return encoding;
    }
    /**
     * Hier wird ein Wert zu den Property <i>encoding</i> gesetzt.
     * Der Typ der Property <i>encoding</i> ist <i>String</i>.
     *
     * @param encoding
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    /**
     * Hier wird ein Wert zu den Property <i>namespace</i> gesetzt.
     * Der Typ der Property <i>namespace</i> ist <i>String</i>.
     *
     * @param namespace
     */
    public void setNamespace(String namespace) {
        this.namespace = namespace;
    }

}
