// $Id: spooler_communication.cxx 4201 2006-03-23 10:16:57Z jz $
/*
    Hier sind implementiert

    Communication::Connection
    Communication
*/


#include "spooler.h"
#include "../kram/sleep.h"
#include "../kram/sos_java.h"

using namespace std;

namespace sos {
namespace spooler {

const int wait_for_port_available = 2*60;   // Soviele Sekunden warten, bis TCP- oder UDP-Port frei wird

#ifdef Z_WINDOWS
#   include <io.h>
    const int EWOULDBLOCK   = 10035;
    const int ENOTSOCK      = 10038;
    const int EADDRINUSE    = WSAEADDRINUSE;
    const int ENOBUFS       = WSAENOBUFS;
#   define ioctl    ioctlsocket
#   define isatty   _isatty
#else
#   include <unistd.h>
#   include <sys/types.h>
#   include <signal.h>
#endif

#ifndef INADDR_NONE
#   define INADDR_NONE (-1)
#endif

//-----------------------------------------------Xml_operation_connection::Xml_operation_connection

Xml_operation_connection::Xml_operation_connection( Communication::Connection* connection )
: 
    Communication::Operation_connection( connection ) 
{
    _indent = connection->_read_socket == STDIN_FILENO;
}

//--------------------------------------------------Xml_operation_connection::connection_lost_event

void Xml_operation_connection::connection_lost_event( const exception* x )
{
    if( _remote_scheduler )
    {
        _remote_scheduler->connection_lost_event( x );
    }
}

//---------------------------------------------------------------------Xml_operation::Xml_operation

Xml_operation::Xml_operation( Xml_operation_connection* operation_connection )
: 
    _zero_(this+1), 
    Operation( operation_connection ),
    _operation_connection( operation_connection )
{
}

//------------------------------------------------------------------Xml_operation::put_request_part

void Xml_operation::put_request_part( const char* data, int length )
{
    _request_is_complete = _xml_end_finder.is_complete( data, length );

    if( length >= 2  &&  data[ length - 2 ] == '\r' )
    {
        _operation_connection->_indent = true;      // CR LF am Ende lsst Antwort einrcken. CR LF soll nur bei telnet-Eingabe kommen.
    }

    _request.append( data, length );
}

//--------------------------------------------------------------------Xml_operation::begin

void Xml_operation::begin()
{
    Command_processor command_processor ( _spooler, _connection->_security_level, this );

    //command_processor.set_communication_operation( this );
    //command_processor.set_host( _connection->peer_host() );

    if( string_begins_with( _request, " " ) )  _request = ltrim( _request );

    _connection->_log.info( message_string( "SCHEDULER-932", _request ) );
    _response = command_processor.execute( _request, Time::now(), _operation_connection->_indent );

    if( _operation_connection->_indent )  _response = replace_regex( _response, "\n", "\r\n" );      // Fr Windows-telnet

    _response += '\0';  // Null-Byte terminiert die XML-Antwort

    if( command_processor._error )  _connection->_log.error( command_processor._error->what() );
}

//----------------------------------------------------Communication::Listen_socket::async_continue_

bool Communication::Listen_socket::async_continue_( Continue_flags flags )
{
    bool something_done = false;

    //if( socket_read_signaled() )
    {
        ptr<Connection> new_connection = Z_NEW( Connection( _communication ) );

        bool ok = new_connection->do_accept( _read_socket );
        if( ok )
        {
            if( _communication->_connection_list.size() >= max_communication_connections )
            {
                _spooler->_log.error( message_string( "SCHEDULER-286", max_communication_connections ) );
            }
            else
            {
                _communication->_connection_list.push_back( new_connection );

                new_connection->add_to_socket_manager( _spooler->_connection_manager );
                new_connection->socket_expect_signals( Socket_operation::sig_read | Socket_operation::sig_write | Socket_operation::sig_except );
            }

            something_done = true;
        }
    }

    async_clear_signaled();
    return something_done;
}

//-------------------------------------------------------Communication::Udp_socket::async_continue_

bool Communication::Udp_socket::async_continue_( Continue_flags )
{
    bool something_done = false;

    //if( socket_read_signaled() )
    {
        char           buffer [4096];
        sockaddr_in    addr;     
        sockaddrlen_t  addr_len = sizeof addr;

        addr.sin_addr.s_addr = 0;
        int len = recvfrom( _read_socket, buffer, sizeof buffer, 0, (sockaddr*)&addr, &addr_len );
        if( len > 0 ) 
        {
            Host host = addr.sin_addr;
            if( _spooler->security_level( host ) < Security::seclev_signal )
            {
                _spooler->log()->error( message_string( "SCHEDULER-301", host.as_string() ) );   //"UDP-Nachricht von " + host.as_string() + " nicht zugelassen." );
            }
            else
            {
                Command_processor command_processor ( _spooler, _spooler->security_level( host ) );
                //command_processor.set_host( host );
                string cmd ( buffer, len );
                _spooler->log()->info( message_string( "SCHEDULER-955", host.as_string(), cmd ) );
                command_processor.execute( cmd, Time::now() );
            }
            
            something_done = true;
        }
    }

    async_clear_signaled();
    return something_done;
}

//------------------------------------------------------------------Communication::Connection::Connection

Communication::Connection::Connection( Communication* communication )
:
    _zero_(this+1),
    _spooler(communication->_spooler),
    _communication(communication),
    _log(_spooler)
{
    //_receive_at_start = true;
    //recv_clear();
}

//-----------------------------------------------------------------Communication::Connection::~Connection

Communication::Connection::~Connection()
{
    remove_from_socket_manager();

    if( _read_socket != SOCKET_ERROR  &&  _read_socket != STDIN_FILENO )
    {
        Z_LOG2( "socket.close", "close(" << _read_socket << ")\n" );
        closesocket( _read_socket );
    }

    _read_socket = SOCKET_ERROR;
    _write_socket = SOCKET_ERROR;
}

//-------------------------------------------------------Communication::Connection::connection_state_name

string Communication::Connection::connection_state_name( Connection_state state )
{
    switch( state )
    {
        case s_none:        return "none";
        case s_ready:       return "ready";
        case s_receiving:   return "receiving";
        case s_processing:  return "processing";
        case s_responding:  return "responding";
        case s_closing:     return "closing";
        default:            return S() << "Connection_state(" << (int)state << ")";
    }
}

//----------------------------------------------------------------Communication::Connection::remove_me

void Communication::Connection::remove_me( const exception* x )
{
    remove_operation();
    if( _operation_connection )  _operation_connection->connection_lost_event( x );

    _communication->remove_connection( this );
}

//----------------------------------------------------------------Communication::Connection::terminate
/*
void Communication::Connection::terminate()
{
    _dont_receive = true;
    async_continue( false );  // Wir warten nicht. Was nicht sofort gesendet werden kann, geht verloren (nur bei sehr vollem Puffer)
}
*/
//----------------------------------------------------------------Communication::Connection::do_accept

bool Communication::Connection::do_accept( SOCKET listen_socket )
{
    try
    {
        bool ok = this->accept( listen_socket );
        if( !ok )  return false;        // EWOULDBLOCK

        //_read_socket.set_linger( true, 0 );
        call_ioctl( FIONBIO, 1 );

        set_buffer_size();

        _security_level = _spooler->security_level( peer_host() );

        _log.set_prefix( "TCP connection to " + _peer_host_and_port.as_string() );


        if( _security_level <= Security::seclev_signal )
        {
            _log.warn( message_string( "SCHEDULER-287" ) );
            do_close();
            return false;
        }

        //set_event_name( S() << "TCP:" << _host.as_string() << ":" << ntohs( _peer_addr.sin_port ) );

        _log.info( message_string( "SCHEDULER-933" ) );
        _connection_state = s_ready;

    }
    catch( const exception& x ) { _log.error(x.what()); return false; }

    return true;
}

//-----------------------------------------------------------------Communication::Connection::do_close

void Communication::Connection::do_close()
{
    remove_operation();
    remove_from_socket_manager();

    if( _read_socket == STDIN_FILENO  &&  _eof  &&  isatty(STDIN_FILENO) )  _spooler->cmd_terminate();  // Ctrl-D (auch wenn Terminal-Sitzung beendet?)

    if( _read_socket != SOCKET_ERROR  &&  _read_socket != STDIN_FILENO )
    {
        Z_LOG2( "socket.close", "close(" << _read_socket << ")\n" );
        closesocket( _read_socket );
    }
    
    _read_socket  = SOCKET_ERROR;
    _write_socket = SOCKET_ERROR;
}

//------------------------------------------------------Communication::Connection::remove_operation

void Communication::Connection::remove_operation()
{
    if( _operation )
    {
        _operation->set_async_manager( NULL );
        _operation->close();
        _operation = NULL;
    }
}

//------------------------------------------------------------------Communication::Connection::do_recv

bool Communication::Connection::do_recv()
{
    bool something_done = false;

    char buffer [ 4096 ];

    int len = call_recv( buffer, sizeof buffer );
    if( len == 0 )  
    {
        something_done = _eof;
    }
    else
    {
        something_done = true;

        const char* p = buffer;

        if( _connection_state == s_ready )   // Das sind die ersten empfangen Bytes der Anforderung
        {
            if( len == 1  &&  buffer[0] == '\x04' )  { _eof = true; return true; }      // Einzelnes Ctrl-D beendet Sitzung

            if( len == 1  &&  buffer[0] == '\n'   
             || len == 2  &&  buffer[0] == '\r'  &&  buffer[1] == '\n' )  { _spooler->signal( "do_something!" );  _spooler->_last_time_enter_pressed = Time::now().as_time_t(); return true; }

            //_receive_at_start = false;
            _connection_state = s_receiving;

            if( !_operation_connection )
            {
                //if( buffer[ 0 ] == '\0' )   // Das erste von vier Lngenbytes ist bei der ersten Nachricht auf jeden Fall 0
                //{
                //    _operation_connection = Z_NEW( Object_server_processor_connection( this ) );
                //}
                //else
                //if( string_begins_with( buffer, "GET /" )  ||  string_begins_with( buffer, "POST /" ) )     // Muss vollstndig im ersten Datenblock enthalten sein!
                if( isupper( (unsigned char)buffer[ 0 ] ) )  // CONNECT, DELETE, GET, HEAD, OPTIONS, POST, PUT, TRACE 
                {
                    _operation_connection = Z_NEW( http::Operation_connection( this ) );
                }
                else
                {
                    _operation_connection = Z_NEW( Xml_operation_connection( this ) );
                }
            }

            _operation = _operation_connection->new_operation();
            _operation->set_async_manager( _spooler->_connection_manager );     // Fr set_async_next_gmtime()
        }

        //if( _read_socket != STDIN_FILENO )  _operation->set_host( &_peer_host_and_port._host );

        if( len > 0 )
        {
            if( z::Log_ptr log = "socket.data" )  *log << "recv: ", log->write( p, len ), *log << endl;
            _operation->put_request_part( p, len );
        }
    }

    return something_done;
}

//----------------------------------------------------------Communication::Connection::async_continue_

bool Communication::Connection::async_continue_( Continue_flags )
{
    //Z_LOG2( "joacim", __FUNCTION__ << state() << "\n" );

    bool something_done = false;

    try
    {
        if( _connection_state == s_responding )  assert_no_recv_data();  //2005-10-25 wegen endloser read-Signalisierung zwischen Firefox und Linux-Scheduler (F5 im Protokollfenster)

        while( _connection_state == s_ready  ||  _connection_state == s_receiving )
        //while( _responding  &&  !_dont_receive )
        {
            bool ok = do_recv();
            if( !ok )  break;

            something_done = true;

            if( _eof )  break;

            if( _operation  &&  _operation->request_is_complete() ) 
            {
                _connection_state = s_processing;
                _operation->begin();
            }
        }

        if( _connection_state == s_processing )
        {
            if( _operation->async_finished() )  _connection_state = s_responding;
        }

        if( _connection_state == s_responding )
        {
            if( state() == s_sending )  something_done |= send__continue();

            while( state() == Buffered_socket_operation::s_ready )
            {
                string data = _operation->get_response_part();
                if( data == "" )  break;

                Z_LOG2( "socket.data", "send: " << data << '\n' );

                something_done = true;

                send__start( data );
            }
                

            if( state() == Buffered_socket_operation::s_ready  &&  _operation->response_is_complete() )
            {
                if( _operation->should_close_connection() )  _eof = true;   // Wir tun so, als ob der Client EOF geliefert hat. Das fhrt zum Schlieen der Verbindung.

                remove_operation();
                _connection_state = s_ready;
            }
        }


        if( _eof )
        { 
            // Bei _eof sofort handeln! Nmlich Verbindung schlieen. 
            // Sonst liefert select() immer wieder den Socket, was zur Schleife fhrt.
            // Andere Lsung: Socket aus read-Menge des select() herausnehmen.

            // _http_response kann bei einem Log endlos sein. Also kein Kriterium, das Schlieen zu verzgern.
            //if( _http_response )  Z_LOG2( "scheduler.http", "Browser schliet Verbindung bevor HTTP-Response fertig gesendet werden konnte\n" );
            remove_me();
            return true; 
        }
    }
    catch( const exception& x ) 
    { 
        if( string_begins_with( x.what(), "ERRNO-103 "     ) == 0                                  // ECONNABORTED, Firefox trennt sich so
         || string_begins_with( x.what(), "WINSOCK-10053 " ) == 0                                  // ECONNABORTED, Firefox trennt sich so
         || string_begins_with( x.what(), "ERRNO-104 "     ) == 0                                  // ECONNRESET, Internet Explorer trennt sich so
         || string_begins_with( x.what(), "WINSOCK-10054 " ) == 0 )  _log.debug( x.what() );       // ECONNRESET, Internet Explorer trennt sich so
                                                               else  _log.error( x.what() );  

        remove_me( &x );
        return true; 
    }

    async_clear_signaled();
    return something_done;
}

//-----------------------------------------------------------Communication::Connection::dom_element

xml::Element_ptr Communication::Connection::dom_element( const xml::Document_ptr& document, const Show_what& what ) const
{
    xml::Element_ptr result = document.createElement( "connection" );

    result.setAttribute( "state", connection_state_name() + "/" + state_name() );


    xml::Element_ptr peer_element = result.append_new_element( "peer" );
    peer_element.setAttribute( "host_ip", peer().host().ip_string() );
    peer_element.setAttribute( "port"   , peer().port() );

    if( peer().host().name() != "" )
    peer_element.setAttribute( "host_name", peer().host().name() );

    if( _operation_connection )
    {
        result.setAttribute( "operation_type", _operation_connection->connection_type() );
    }

    if( _operation )
    {
        //result.setAttribute( "operation", _operation->obj_name() );
        xml::Element_ptr operation_element = result.append_new_element( "operation" );
        
        double gmtimeout_at = _operation->gmtimeout();
        if( gmtimeout_at < double_time_max )  operation_element.setAttribute( "timeout_at", Time( localtime_from_gmtime( gmtimeout_at ) ).as_string() );

        operation_element.appendChild( _operation->dom_element( document, what ) );
    }

    return result;
}

//--------------------------------------------------------------------Communication::Communication

Communication::Communication( Spooler* spooler )
: 
    _zero_(this+1), 
    _spooler(spooler),
    _listen_socket(this),
    _udp_socket(this)
{
}

//--------------------------------------------------------------------Communication::~Communication

Communication::~Communication()
{
    close( 0.0 );

    if( _initialized ) 
    {
#       ifdef SYSTEM_WIN
            WSACleanup();
#       endif
    }
}

//-----------------------------------------------------------------------------Communication::close

void Communication::close( double wait_time )
{
    // Funktioniert nicht: double until = Time::now() + wait_time;

    _listen_socket.set_linger( true, 0 );
    _listen_socket.close();

    _udp_socket.set_linger( true, 0 );
    _udp_socket.close();


    //while(1)
    {
        //bool responding = false;

        Connection_list::iterator c = _connection_list.begin();
        while( c != _connection_list.end() )
        {
            ptr<Connection> connection = *c;
            c++;
            
            connection->set_linger( true, 1 );     // 1: Eine Sekunde Zeit, um asynchron (non-blocking) die Verbindung abzubauen. sleep(2) in spooler.cxx!
                                                // 0: close() bricht Verbindung ab (jedenfalls unter Unix). Schickt RST. D
                                                // Damit bleibt beim Neustart Windows-Schedulers der Browser nicht kleben (der kriegt den Verbindungsabbau nicht mit)

            //connection->terminate();    // Kann Connection aus _connection_list entfernen.
            connection->_connection_state = Connection::s_closing;
            connection->async_continue();
            //if( connection->_responding )
            //{
            //    responding = true;
            //    Z_LOG( "Warten auf " << *connection << '\n' );
            //}
        }

        //if( !responding )  break;
        //if( Time::now() > until )  break;

        //sleep( 0.1 );
    }


    //c = _connection_list.begin();
    //while( c != _connection_list.end() )  remove_connection( *c );

    _connection_list.clear();

    _terminate = true;
}

//--------------------------------------------------------------------Communication::remove_connection

void Communication::remove_connection( Connection* connection )
{
    FOR_EACH( Connection_list, _connection_list, it )
    {
        if( *it == connection )  
        {
            ptr<Connection> c = *it;
            _connection_list.erase( it ); 
            c->do_close(); 
            break;
        }
    }
}

//----------------------------------------------------------------Communication::main_thread_exists
/*
bool Communication::main_thread_exists()
{
#   ifndef Z_LINUX

        return true;

#    else

        if( kill( _spooler->_pid, 0 ) == -1  &&  errno == ESRCH )
        {
            //_spooler->_log.error( "Kommunikations-Thread wird beendet, weil der Hauptthread (pid=" + as_string(_spooler->_pid) + ") verschwunden ist" );
            //_spooler->_log verklemmt sich manchmal nach Ausgabe des Zeitstemples (vor "[ERROR]"). Vielleicht wegen einer Semaphore?

            //fprintf( stderr, "Kommunikations-Thread wird beendet, weil der Hauptthread (pid=%d) verschwunden ist\n", _spooler->_pid );
            return false;  //?  Thread bleibt sonst hngen, wenn Java sich bei Ctrl-C sofort verabschiedet. Java lsst SIGINT zu, dieser Thread aber nicht.
        }

        return true;

#   endif
}
*/
//-----------------------------------------------------------------------Communication::bind_socket

int Communication::bind_socket( SOCKET socket, struct sockaddr_in* sa, const string& tcp_or_udp )
{
    int ret;
    int true_ = 1;

    string port_name = tcp_or_udp + "-port " + as_string( ntohs( sa->sin_port ) );

    if( _spooler->_reuse_port )     // War fr Suse 8 ntig. Und fr Windows XP, wenn Scheduler zuvor abgestrzt ist (mit Debugger), denn dann bleibt der Port ewig blockiert
    {
        _spooler->_log.warn( message_string( "SCHEDULER-288", port_name ) );
        LOG( "setsockopt(" << socket << ",SOL_SOCKET,SO_REUSEADDR,1)  " );
        ret = setsockopt( socket, SOL_SOCKET, SO_REUSEADDR, (const char*)&true_, sizeof true_ );
        LOG( "ret=" << ret );  if( ret == SOCKET_ERROR )  LOG( "errno=" << errno << "  "  << z_strerror(errno) );
        LOG( "\n" );
    }

    bool print_dots = isatty( fileno(stderr) ) && isatty( fileno(stdin) );

    ret = ::bind( socket, (struct sockaddr*)sa, sizeof (struct sockaddr_in) );

    if( ret == SOCKET_ERROR  &&  socket_errno() == EADDRINUSE )
    {
        int my_errno = errno;  // Nur fr Unix

        _spooler->_log.warn( message_string( "SCHEDULER-289", port_name, wait_for_port_available ) );

        for( int i = 1; i <= wait_for_port_available; i++ )
        {
            if( ctrl_c_pressed || _spooler->state() == Spooler::s_stopping || _spooler->state() == Spooler::s_stopped )  { errno = EINTR; return SOCKET_ERROR; }
            //if( !main_thread_exists() )  return EINTR;  //?  Thread bleibt sonst hngen, wenn Java sich bei Ctrl-C sofort verabschiedet. Java lsst SIGINT zu, dieser Thread aber nicht.

            sos_sleep(1);
        
            ret = ::bind( socket, (struct sockaddr*)sa, sizeof (struct sockaddr_in) );
            my_errno = errno;
            if( ret != SOCKET_ERROR ) break;
            if( socket_errno() != EADDRINUSE )  break;
            if( ctrl_c_pressed )  break;

            if( print_dots )  fputc( i % 10 == 0? '0' + i / 10 % 10 : '.', stderr );
        }

        if( print_dots )  fputc( '\n', stderr );

        errno = my_errno;   // Nur fr Unix
    }

    return ret;
}

//------------------------------------------------------------------------------Communication::bind

void Communication::bind()
{
    struct sockaddr_in  sa;
    int                 ret;
    unsigned long       on = 1;


    // UDP:

    if( _udp_port != _spooler->udp_port() )
    {
        //THREAD_LOCK( _semaphore )
        {
            _udp_socket.close();
            _udp_port = 0;

            if( _spooler->udp_port() != 0 )
            {
                _udp_socket._read_socket = socket( AF_INET, SOCK_DGRAM, 0 );
                if( _udp_socket._read_socket == SOCKET_ERROR )  throw_socket( socket_errno(), "socket" );

                //_udp_socket.set_linger( false );
                
                sa.sin_port        = htons( _spooler->udp_port() );
                sa.sin_family      = AF_INET;
                sa.sin_addr.s_addr = 0; // INADDR_ANY

                ret = bind_socket( _udp_socket._read_socket, &sa, "UDP" );
                if( ret == SOCKET_ERROR )  throw_socket( socket_errno(), "udp-bind ", as_string(_spooler->udp_port()).c_str() );

                ret = ioctlsocket( _udp_socket._read_socket, FIONBIO, &on );
                if( ret == SOCKET_ERROR )  throw_socket( socket_errno(), "ioctl(FIONBIO)" );

                _udp_port = _spooler->udp_port();
                _rebound = true;

                _udp_socket.set_event_name( S() << "UDP:" << ntohs( sa.sin_port ) );

                _spooler->log()->info( message_string( "SCHEDULER-956", "UDP", _udp_port ) );    // "Scheduler erwartet Kommandos ber UDP-Port " + sos::as_string(_udp_port) );
            }
        }

        _udp_socket.add_to_socket_manager( _spooler->_connection_manager );
        _udp_socket.socket_expect_signals( Socket_operation::sig_read );
    }


    // TCP: 

    if( _tcp_port != _spooler->tcp_port() )
    {
        //Z_MUTEX( _semaphore )
        {
            _listen_socket.close();
            _tcp_port = 0;

            if( _spooler->tcp_port() != 0 )
            {
                _listen_socket.set_event_name( S() << "TCP listen(" << _spooler->tcp_port() << ")" );

                _listen_socket._read_socket = socket( AF_INET, SOCK_STREAM, 0 );
                if( _listen_socket._read_socket == SOCKET_ERROR )  throw_socket( socket_errno(), "socket" );

                //_listen_socket.set_linger( false );
                
                sa.sin_port        = htons( _spooler->tcp_port() );
                sa.sin_family      = AF_INET;
                sa.sin_addr.s_addr = 0; // INADDR_ANY

                ret = bind_socket( _listen_socket._read_socket, &sa, "TCP" );
                if( ret == SOCKET_ERROR )  throw_socket( socket_errno(), "tcp-bind", as_string(_spooler->tcp_port()).c_str() );

                Z_LOG2( "socket.listen", "listen()\n" );
                ret = listen( _listen_socket._read_socket, 5 );
                if( ret == SOCKET_ERROR )  throw_socket( socket_errno(), "listen" );

                ret = ioctlsocket( _listen_socket._read_socket, FIONBIO, &on );
                if( ret == SOCKET_ERROR )  throw_socket( socket_errno(), "ioctl(FIONBIO)" );

                _tcp_port = _spooler->tcp_port();
                _rebound = true;

                _spooler->log()->info( message_string( "SCHEDULER-956", "TCP", _tcp_port ) );    // "Scheduler erwartet Kommandos ber $1-Port " 
            }
        }

        _listen_socket.add_to_socket_manager( _spooler->_connection_manager );
        _listen_socket.socket_expect_signals( Socket_operation::sig_read );
    }


#   ifndef Z_WINDOWS
        if( _spooler->_interactive )
        {
            ret = ioctl( STDIN_FILENO, FIONBIO, &on );   // In Windows nicht mglich
            if( ret == 0 )
            {
                ptr<Connection> new_connection = Z_NEW( Connection( this ) );
        
                new_connection->_read_socket  = STDIN_FILENO;
                new_connection->_write_socket = STDOUT_FILENO;
                new_connection->_socket_send_buffer_size = 1024;

                new_connection->add_to_socket_manager( _spooler->_connection_manager );
                new_connection->socket_expect_signals( Socket_operation::sig_read );

                new_connection->set_event_name( "stdin" );

                _connection_list.push_back( new_connection );
            }
        }
#   endif
}

//-----------------------------------------------------------------------------Communication::init

void Communication::init()
{
    if( _initialized )  return;

#   ifdef SYSTEM_WIN
        WSADATA wsa_data;
        int ret = WSAStartup( 0x0101, &wsa_data );
        if( ret )  throw_socket( ret, "WSAStartup" );
#   endif

    _initialized = true;
}

//-------------------------------------------------------------------Communication::start_or_rebind

void Communication::start_or_rebind()
{
    if( !_started )  
    {
        init();
        bind();
    }
    else
    {
        rebind();
    }
}

//-----------------------------------------------------------------------Communication::dom_element

xml::Element_ptr Communication::dom_element( const xml::Document_ptr& document, const Show_what& what ) const
{
    xml::Element_ptr connections_element = document.createElement( "connections" );

    Z_FOR_EACH_CONST( Connection_list, _connection_list, c )
    {
        const Connection* connection = *c;

        connections_element.appendChild( connection->dom_element( document, what ) );
    }

    return connections_element;
}

//-------------------------------------------------------------------------------------------------

} //namespace spooler
} //namespace sos

