// $Id: java.cxx 4173 2006-03-09 13:58:17Z jz $

#include "zschimmer.h"
#include "threads.h"
#include "z_com.h"
#include "java.h"

#ifdef Z_WINDOWS
#   include "z_windows_registry.h"          // Der Pfad der jvm.dll steht in der Registry
#endif

#ifdef Z_UNIX
#   include <dlfcn.h>
#endif

#ifdef Z_HPUX
#   include <dl.h>
#endif


namespace zschimmer {
namespace java {

using namespace std;

//--------------------------------------------------------------------------------------------const

static Message_code_text error_codes[] =
{
    { "Z-JAVA-100", "Die Java Virtual Machine kann nicht geladen werden." },
    { "Z-JAVA-101", "Die Java Virtual Machine ist nicht gestartet" },
    { "Z-JAVA-102", "Fehler $1 in Java aufgetreten bei $2" },
    { "Z-JAVA-103", "Der Eintrag [java] vm= fehlt in der sos.ini" },
    { "Z-JAVA-104", "Variant-Typ $1 kann nicht in einen Java-Typ konvertiert werden" },
    { "Z-JAVA-105", "Java-Exception $1(\"$2\"), method=$3" },
    { "Z-JAVA-106", "NULL-Object" },
    { "Z-JAVA-107", "Typ '$1' erwartet, ist aber '$2'" },
    { "Z-JAVA-109", "set_jvm: Eine Java Virtual Machine ist bereits gestartet" },
    { "Z-JAVA-110", "Die Java Virtual Machine ist lter als Version 1.2" },
    { "Z-JAVA-111", "Objekt-Verweis ist nicht mehr gltig" },
    { "Z-JAVA-112", "Der $1. Parameter ist von unbekanntem Typ" },
    { "Z-JAVA-113", "Name ist zu lang: $1" },
    {}
};

//-------------------------------------------------------------------------------------------static

Vm*                             Vm::static_vm         = NULL;

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( java )
{
    add_message_code_texts( error_codes );
}

//------------------------------------------------------------------------------string_from_jstring

string string_from_jstring( JNIEnv* jenv, const jstring& jstr )
{
    if( !jstr )  return "";

    return com::string_from_ole( (const OLECHAR*)(const jchar*)Read_jstring_critical( jenv, jstr ), jenv->GetStringLength( jstr ) );
}

//------------------------------------------------------------------------------jstring_from_string

jstring jstring_from_string( JNIEnv* jenv, const char* str, size_t length )
{
    com::Bstr bstr ( str, length );
    return jenv->NewString( (const jchar*)(const OLECHAR*)bstr, bstr.length() );
}

//-------------------------------------------------------------------------------set_java_exception

void set_java_exception( JNIEnv* jenv, const char* what )
{
    const char* exception_class_name = "java/lang/RuntimeException";
    jclass exception_class = jenv->FindClass( exception_class_name  );

    if( exception_class )  
    {
        int err = jenv->ThrowNew( exception_class, what ); 
        jenv->DeleteLocalRef( exception_class );

        if( err == 0 )  return;
    }

    string w = what + string(" --- JNI.FindClass() liefert nicht Klasse ") + exception_class_name;

    if( jenv->ExceptionCheck() )  jenv->ExceptionDescribe();      // schreibt nach stderr

    jenv->FatalError( w.c_str() );
}

//-------------------------------------------------------------------------------set_java_exception

void set_java_exception( JNIEnv* jenv, const exception& x )
{
    set_java_exception( jenv, x.what() );
}

//-------------------------------------------------------------------------------set_java_exception

void set_java_exception( JNIEnv* jenv, const _com_error& x ) 
{
    string what = com::string_from_ole( x.Description() );
    set_java_exception( jenv, what.c_str() );
}

//-------------------------------------------------------------------------set_NullPointerException

void set_NullPointerException( JNIEnv* jenv, const string& what )
{
    const char* exception_class_name = "java/lang/NullPointerException";
    jclass exception_class = jenv->FindClass( exception_class_name  );

    if( exception_class )  
    {
        int err = jenv->ThrowNew( exception_class, what.c_str() ); 
        jenv->DeleteLocalRef( exception_class );

        if( err == 0 )  return;
    }

    string w = what + string(" --- JNI.FindClass() liefert nicht Klasse ") + exception_class_name;

    if( jenv->ExceptionCheck() )  jenv->ExceptionDescribe();      // schreibt nach stderr

    jenv->FatalError( w.c_str() );
}

//-----------------------------------------------------------------------------------get_class_name

string get_class_name( JNIEnv* jenv, jclass c )
{
    jmethodID getClass_method = jenv->GetMethodID( c, "getClass", "()Ljava/lang/Class;" );
    if( getClass_method == 0 )  Env(jenv).throw_java( "getMethodID()" );

    Local_jobject cls ( Vm::static_vm, jenv->CallObjectMethod( c, getClass_method ) );

    jstring result_j = (jstring)jenv->CallObjectMethod( c, jenv->GetMethodID( cls.get_jclass(), "getName", "()Ljava/lang/String;" ) );
    
    string result = string_from_jstring( jenv, result_j );
   
    jenv->DeleteLocalRef( result_j );

    return result;
}

//-----------------------------------------------------------------------------------get_class_name

string get_class_name( JNIEnv* jenv, jobject jobj )
{
    return get_class_name( jenv, jenv->GetObjectClass( jobj ) );
}

//--------------------------------------------------------------------------------------Vm::set_jvm
/*
void Vm::set_jvm( JavaVM* jvm )
{
    Z_MUTEX( static_vm_mutex )
    {
    	if( jvm )
    	{
            if( static_vm )
            {
                if( jvm == static_vm->_vm )  return;
                throw_xc( "Z-JAVA-109" );
            }

            ptr<Vm> my_vm = Z_NEW( Vm( jvm ) );
            static_foreign_vm = my_vm.take();
        }
        else
        {
            if( static_foreign_vm )
            {
                static_foreign_vm->Release();
                static_foreign_vm = NULL;
            }
        }
    }
}
*/
//---------------------------------------------------------------------------------------Vm::get_vm

ptr<Vm> Vm::get_vm( bool start )
{
    ptr<Vm> vm;

    Z_MUTEX( static_vm_mutex )
    {
        if( static_vm )
        {
            vm = static_vm;
        }
        else
        {
            vm = Z_NEW( Vm( false ) );
        }

        if( start )  vm->start();
    }

    return vm;
}

//----------------------------------------------------------------------------------Vm::set_options

void Vm::set_options( const string& options )
{
    const char* p = options.c_str();
    while( *p == ' ' )  p++;

    while( *p )
    {
        string option;
        const char* blank = strchr( p, ' ' );
        if( !blank )  blank = p + strlen(p);

        const char* q = strchr( p, '=' );

        if( q )  q++;
           else  q = p + strlen(p);

        if( q < blank )
        {
            // key=value, key="value", key='value'

            option = string( p, q - p );

            if( !*q )  break;

            p = q;
            if( *p == '"' || *p == '\'' )
            {
                char quote = *p++;
                while( *p  &&  *p != quote )
                {
                    if( *p == '\\' )  p++;
                    option += *p++;
                }

                if( *p == quote )  p++;
            }
            else
            {
                q = strchr( p, ' ' );
                if( !q )  q = p + strlen(p);
                option.append( p, q - p );
                p = q;
            }
        }
        else 
        {
            option = string( p, blank - p );
            p = blank;
        }

        while( *p == ' ' )  p++;

        _options.push_back( option );
    }
}

//------------------------------------------------------------------------------------java_vfprintf

static jint JNICALL java_vfprintf( FILE*, const char *format, va_list args )
{
  //static string   line;
    char            buf[1024];
    int             ret;

#   ifdef Z_WINDOWS
        ret = _vsnprintf_s( buf, sizeof(buf), _TRUNCATE, format, args );
#   else
        memset( buf, 0, sizeof buf );
        ret = _vsnprintf( buf, sizeof(buf), format, args );
        buf[ sizeof buf - 1 ] = '\0';
#   endif

    if( Vm::static_vm ) 
    {
        if( strcmp( buf, "\n" ) == 0 )  return ret;     // log macht sowieso einen Zeilenwechsel, das hier gbe also nur eine Leerzeile

        Vm::static_vm->_log.debug3( buf );
    }
    else
    {
        Z_LOG( buf );
        fputs( buf, stderr );
    }

    return ret;
}

//-------------------------------------------------------------------------------------------Vm::Vm

Vm::Vm( bool do_start )
: 
    _zero_(this+1) 
{ 
    init();

    if( do_start )  start(); 
}

//-------------------------------------------------------------------------------------------Vm::Vm

Vm::Vm( JavaVM* jvm )
: 
    _zero_(this+1),
    _foreign(true)
{ 
    init();

    _vm = jvm;
    _foreign = true;

    JNIEnv* jenv = NULL;
    int ret = _vm->GetEnv( (void**)&jenv, JNI_VERSION_1_2 );
    if( ret != JNI_OK )  throw_xc( "Z-JAVA-110" );

    //log_version( jenv, "" );
}

//------------------------------------------------------------------------------------------Vm::~Vm

Vm::~Vm()
{
    try
    {
        close(); 
    }
    catch( const exception& x ) { Z_LOG( "java::~Vm(): " << x << "\n" ); }

    _options.clear();
    delete _vm_args.options;  _vm_args.options = NULL;
}

//-----------------------------------------------------------------------------------------Vm::init

void Vm::init()
{
    Z_MUTEX( static_vm_mutex )
    {
        if( static_vm )  throw_xc( "Z-JAVA-109" );
        static_vm = this;
    }

    _log.set_prefix( "Java" );
}

//----------------------------------------------------------------------------------Vm::log_version

void Vm::log_version( JNIEnv* jenv, const string& module_filename )
{
    int version = jenv->GetVersion();
    Z_LOG( "JNI " + as_string( version >> 16 ) + "." + as_string( version & 0xFFFF ) + " " + module_filename + " geladen\n" );
}

//-------------------------------------------------------------------------------------set_filename

void Vm::set_filename()
{
#   ifdef Z_WINDOWS

        windows::Registry_key hkey;
        windows::Registry_key version_hkey;

        if( hkey.try_open( HKEY_LOCAL_MACHINE, "software\\JavaSoft\\Java Runtime Environment", KEY_QUERY_VALUE ) )
        {
            string current_version = hkey.get_string( "CurrentVersion", "" );
            if( current_version != "" 
             && version_hkey.try_open( hkey, current_version, KEY_QUERY_VALUE ) )
            {
                _filename = version_hkey.get_string( "RuntimeLib", "" );
            }
        }

        if( _filename == "" )  _filename = "jvm.dll";

#   elif defined Z_HPUX

        _filename = "libjvm.sl";

#   else

        _filename = "libjvm.so";

#   endif
}

//----------------------------------------------------------------------------------------Vm::start
// Im Haupt-Thread zu rufen.

void Vm::start()
{
    if( _javac_filename == "" )  _javac_filename = "javac";

    Z_MUTEX( static_vm_mutex )
    {
        if( !_vm )
        {
            if( com::static_com_context_ptr )
            {
                _vm = request_jvm();
                _foreign = true;
                _vm_requested = true;
            }
/*
            else
            if( static_foreign_jvm )        // veraltet (noch fr hostjava -> hostole)
            {
                _vm = static_foreign_jvm;
                _foreign = true;
            }
*/
        }
    }

    if( _vm )  
    {
        JNIEnv* jenv = NULL;
        int ret = _vm->GetEnv( (void**)&jenv, JNI_VERSION_1_2 );
        if( ret != JNI_OK )  throw_xc( "Z-JAVA-110" );

        return;
    }


    int ret;

    _debug = log_category_is_set( "java" );


    if( _filename == "" )  
    {
        set_filename();
    }

    string module_filename = _filename;

    typedef int JNICALL JNI_CreateJavaVM_func( JavaVM**, JNIEnv**, JavaVMInitArgs* );

    JNI_CreateJavaVM_func* JNI_CreateJavaVM;

#   ifdef Z_WINDOWS
    {    
        Z_LOG( "LoadLibrary " << module_filename << '\n' );
        HINSTANCE vm_module = LoadLibrary( module_filename.c_str() );
        if( !vm_module )  throw_mswin( "LoadLibrary", "Java Virtual Machine " + module_filename );

        //module_filename = filename_of_hinstance( vm_module );
        //Z_LOG( "HINSTANCE=" << (void*)vm_module << "  " << module_filename << "  " << file_version_info( module_filename ) << '\n' );

        JNI_CreateJavaVM = (JNI_CreateJavaVM_func*)GetProcAddress( vm_module, "JNI_CreateJavaVM" );
        if( !JNI_CreateJavaVM )  throw_mswin( "GetProcAddress", "JNI_CreateJavaVM" );
    }
#   else
    {
        Z_LOG( "dlopen " << module_filename << ",RTLD_LAZY\n" );
        void *vm_module = dlopen( module_filename.c_str(), RTLD_LAZY );
        if( !vm_module )  throw_xc( "Z-JAVA-100", dlerror(), module_filename.c_str() );

        JNI_CreateJavaVM = (JNI_CreateJavaVM_func*)dlsym( vm_module, "JNI_CreateJavaVM" );
        if( !JNI_CreateJavaVM )  throw_xc( "Z-JAVA-100", dlerror(), "JNI_CreateJavaVM" );
    }
#   endif

  //if( _work_class_dir != ""    )  _complete_class_path += _work_class_dir + Z_PATH_SEPARATOR;
    if( _class_path != "" )  _complete_class_path += _class_path + Z_PATH_SEPARATOR;
    _complete_class_path = replace_regex( _complete_class_path, Z_PATH_SEPARATOR "$", "" );
    _complete_class_path = replace_regex( _complete_class_path, Z_PATH_SEPARATOR "+", Z_PATH_SEPARATOR );


    _vm_args.version = JNI_VERSION_1_2;
    _vm_args.ignoreUnrecognized = true;

    // _options.push_back( "-Xfuture" );   
    // Performs strict class-file format checks.  For purposes of backwards compatibility, the default format checks performed by the Java 2 SDK's virtual machine  are
    // no  stricter  than the checks performed by 1.1.x versions of the JDK software.  The -Xfuture flag turns on stricter class-file format checks that enforce closer
    // conformance to the class-file format specification.  Developers are encouraged to use this flag when developing new code because the stricter checks will become
    // the default in future releases of the Java application launcher.
    

    if( _debug ) 
    {
        _options.push_back( Option( "-Xdebug"               ) );
        _options.push_back( Option( "-showversion"          ) );
        _options.push_back( Option( "-verbose:class,gc,jni" ) );
        _options.push_back( Option( "-Xcheck:jni "          ) );
    }

    _options.push_back( Option( "-Xfuture"        ) );  // Perform strict class-file format checks.
    _options.push_back( Option( "-XX:+UseAltSigs" ) );  // Statt SIGUSR1 und SIGUSR2 soll Java andere Signale (welche?) benutzen.
  //_options.push_back( Option( "-Xnocatch"       ) );  // disable the catch all signal handler

//#   ifdef Z_UNIX
    _options.push_back( Option( "-Xrs"            ) );  // signal masks for SIGINT, SIGTERM, SIGHUP, and SIGQUIT are not changed by the JVM, and signal handlers for these signals are not installed
    /*
        -Xrs 
        Reduces use of operating-system signals by the Java virtual machine (JVM). 
        In a previous release, the Shutdown Hooks facility was added to allow orderly shutdown of a Java application.
        The intent was to allow user cleanup code (such as closing database connections) to run at shutdown, even if the JVM terminates abruptly. 

        Sun's JVM catches signals to implement shutdown hooks for abnormal JVM termination. 
        The JVM uses SIGHUP, SIGINT, and SIGTERM to initiate the running of shutdown hooks. 

        The JVM uses a similar mechanism to implement the pre-1.2 feature of dumping thread stacks for debugging purposes. 
        Sun's JVM uses SIGQUIT to perform thread dumps. 

        Applications embedding the JVM frequently need to trap signals like SIGINT or SIGTERM, which can lead to interference 
        with the JVM's own signal handlers. The -Xrs command-line option is available to address this issue. When -Xrs is used on 
        Sun's JVM, the signal masks for SIGINT, SIGTERM, SIGHUP, and SIGQUIT are not changed by the JVM, and signal handlers for
        these signals are not installed. 

        There are two consequences of specifying -Xrs: 

        SIGQUIT thread dumps are not available. 
        User code is responsible for causing shutdown hooks to run, for example by calling System.exit() when the JVM is to be terminated.
    */
//#   endif

    _options.push_back( Option( "vfprintf", (void*)java_vfprintf ) );
    _options.push_back( "-Djava.class.path=" + _complete_class_path );

    _vm_args.nOptions = (int)_options.size();
    delete _vm_args.options;
    _vm_args.options = new JavaVMOption[ _vm_args.nOptions ];

    Z_LOG( "Java options: " );
    for( int i = 0; i < _vm_args.nOptions; i++ )
    {
        _vm_args.options[i].optionString = (char*)_options[i]._option.c_str();
        _vm_args.options[i].extraInfo    =        _options[i]._extra;
        Z_LOG( _vm_args.options[i].optionString << "  " );
    }
    Z_LOG( '\n' );

    Z_LOG( "JNI_CreateJavaVM()\n" );
    JNIEnv* jenv = NULL;
    ret = JNI_CreateJavaVM( &_vm, &jenv, &_vm_args );

    //if( static_vm == this )  set_jvm( _vm );

  //Z_LOG( "setlocale(LC_ALL,\"C\")\n" );
    const char* java_locale = setlocale( LC_ALL, "C" );
    if( strcmp( java_locale, "C" ) != 0 )  Z_LOG( "Javas locale war " << java_locale << "\n" );

    if( ret < 0 )  throw_java_ret( ret, "JNI_CreateJavaVM", module_filename );

    log_version( jenv, module_filename );

    new_instances( _new_instances );
}

//----------------------------------------------------------------------------------------Vm::close

void Vm::close()
{
    if( _vm )  
    {
        if( _foreign )
        {
            if( _vm_requested )  release_jvm( _vm );
            _vm = NULL;
            Z_MUTEX( static_vm_mutex )  static_vm = NULL;
        }
        else
        {
            if( _dont_destroy  ||  unloading_module )
            {
                Z_LOG( "DestroyJavaVM() unterdrckt\n" );
                //_vm = NULL;
            }
            else
            {
                Env e = env();

                Z_FOR_EACH( Instances, _instances, i )  if( i->_jobject )  e->DeleteGlobalRef( i->_jobject );
                _instances.clear();

                Z_LOG( "DestroyJavaVM() ...\n" );

                int ret = _vm->DestroyJavaVM();
                Z_MUTEX( static_vm_mutex )  static_vm = NULL;

                Z_LOG( "DestroyJavaVM()  OK\n" );            

                if( ret < 0 )  _log.error( "DestroyJavaVM() liefert " + as_string(ret) + ". Java lsst sich nicht entladen" );
                         else  _vm = NULL;
            }

            //if( static_vm == this )  set_jvm( _vm );
        }
    }
}

//-------------------------------------------------------------------------------Vm::set_destroy_vm

void Vm::set_destroy_vm( bool destroy )
{ 
    if( !destroy  &&  !_dont_destroy )  AddRef();       // Nie freigeben, static_vm bleibt erhalten fr erneute Verwendung
    if( destroy   &&   _dont_destroy )  Release();      // AddRef() wieder aufheben

    _dont_destroy = !destroy; 
}

//--------------------------------------------------------------------------------Vm::new_instances

void Vm::new_instances( const string& class_names_str )
{
    if( class_names_str.empty() )  return;

    if( !_vm )
    {
        _new_instances += " ";
        _new_instances += class_names_str;
        return;
    }


    Env            e               = env();
    Class          java_lang_class ( this, "java/lang/Class" );
    vector<string> class_names     = zschimmer::vector_split( " *[ :;,] *", class_names_str );

    for( vector<string>::iterator i = class_names.begin(); i != class_names.end(); i++ )
    {
        string class_name = *i;

        if( !class_name.empty() )
        {
            bool old = false;
            Z_FOR_EACH( Instances, _instances, j )  if( class_name == j->_class_name )  { old = true; break; }
            if( !old )
            {
                Local_jobject cls ( this );

                try
                {
                    Z_LOG( "java Class(\"" << class_name << "\").newInstance()\n" );

                    Class c ( this, class_name ); 
                    jobject obj = e->CallObjectMethod( c, java_lang_class.method_id( "newInstance", "()Ljava/lang/Object;" ) );

                    //jobject obj = e->CallObjectMethod( c, e->GetMethodID( c, "newInstance", "()Ljava/lang/Object;" ) );
    /*
                    Z_LOG( "java.lang.Class.forName(\"" << class_name << "\").newInstance()\n" );

                    cls = e->CallStaticObjectMethod( java_lang_class,
                                                     java_lang_class.static_method_id( "forName", "(Ljava/lang/String;)Ljava/lang/Class;" ),
                                                     (jstring)Jstring( this, class_name ) );
                    if( !cls ) e.throw_java( "Class.forName", class_name );

                    jobject obj = e->CallObjectMethod( cls, java_lang_class.method_id( "newInstance", "()Ljava/lang/Object;" ) );
    */
                    if( !obj )  e.throw_java( "newInstance", class_name );

                    _instances.push_back( Instance( class_name, e->NewGlobalRef( obj ) ) );
                }
                catch( const exception& x ) 
                { 
                    _log.warn( "Java-Klasse \"" + class_name + "\": " + x.what() );
                }
            }
        }
    }
}

//--------------------------------------------------------------------------------Vm::attach_thread

JNIEnv* Vm::attach_thread( const string& thread_name )
{
    if( !_vm )  throw_xc( "Z-JAVA-101" );


    JavaVMAttachArgs args;
    JNIEnv*          jenv = NULL;

    memset( &args, 0, sizeof args );
    args.version = JNI_VERSION_1_2;
    args.name    = (char*)thread_name.c_str();

    if( _debug )  Z_LOG( "java AttachCurrentThread()\n" );

    int ret = _vm->AttachCurrentThread( (void**)&jenv, &args ); 
    if( ret < 0 )  throw_java_ret( ret, "AttachCurrentThread" );

    return jenv;
}

//--------------------------------------------------------------------------------Vm::detach_thread

void Vm::detach_thread()
{
    JNIEnv* jenv = NULL;

    if( _vm  &&  _vm->GetEnv( (void**)&jenv, JNI_VERSION_1_2 ) == JNI_OK )
    {
        if( _debug )  Z_LOG( "java DetachCurrentThread()\n" );

        int ret = _vm->DetachCurrentThread(); 
        if( ret < 0 )  throw_java_ret( ret, "DetachCurrentThread" );
    }
}

//-----------------------------------------------------------------------------------------Vm::jenv

JNIEnv* Vm::jenv() 
{
    if( !_vm )  throw_xc( "Z-JAVA-101" );


    JNIEnv* jenv = NULL;

    if( _vm->GetEnv( (void**)&jenv, JNI_VERSION_1_2 ) == JNI_OK )  return jenv;
                                                             else  return attach_thread( "" );
}

//-----------------------------------------------------------------------------------------Env::Env

Env::Env( Vm* vm )
:
    _jenv( vm->jenv() )
{
}

//-----------------------------------------------------------------------------------throw_java_ret
                                                                                                   
Z_NORETURN void throw_java_ret( int return_value, const string& text1, const string& text2 )
{
    string ret_text;

    switch( return_value )
    {
        case JNI_OK:        ret_text = "ret=JNI_OK";                                            break;
        case JNI_ERR:       ret_text = "ret=JNI_ERR";                                           break;
        case JNI_EDETACHED: ret_text = "ret=JNI_EDETACHED \"thread detached from the VM\"";     break;
        case JNI_EVERSION:  ret_text = "ret=JNI_EVERSION \"JNI version error\"";                break;

#     ifdef JNI_VERSION_1_4
        case JNI_ENOMEM:    ret_text = "ret=JNI_ENOMEM \"not enough memory\"";                  break;
        case JNI_EEXIST:    ret_text = "ret=JNI_EEXIST \"VM already created\"";                 break;
        case JNI_EINVAL:    ret_text = "ret=JNI_EINVAL \"invalid arguments\"";                  break;
#     endif

        default:            ret_text = "ret=" + as_string(return_value);
    }

    Xc xc ( "Z-JAVA-102" );

    xc.insert( 1, ret_text );
    xc.append_text( text1 );
    xc.append_text( text2 );

    throw_xc( xc );
}

//----------------------------------------------------------------------------------Env::throw_java

void Env::throw_java( const string& text1, const string& text2 )
{
    JNIEnv* jenv = this->jenv();

    Local_frame local_frame ( jenv, 10 );
    Java_exception xc ( "Z-JAVA-105" );
    
    xc._exception_name = "Unbekannt";

    if( jenv )
    {
        jthrowable x = jenv->ExceptionOccurred();
        if( x )
        {
            Z_LOG( "Vm::static_vm=" << (void*)Vm::static_vm << "  _debug=" << ( Vm::static_vm? Vm::static_vm->_debug : false ) << "\n" );
            if( Vm::static_vm && Vm::static_vm->_debug )  { Z_LOG( "jenv->ExceptionDescribe()\n" ); jenv->ExceptionDescribe(); }     // Ausgabe ber java_vfprintf()
            jenv->ExceptionClear();

            jclass c = jenv->GetObjectClass(x);
            if( c ) 
            {
                Local_jobject cls ( Vm::static_vm, jenv->CallObjectMethod( x, jenv->GetMethodID( c, "getClass", "()Ljava/lang/Class;" ) ) );
                xc._exception_name = string_from_jstring( (jstring)jenv->CallObjectMethod( cls, jenv->GetMethodID( jenv->GetObjectClass(cls), "getName", "()Ljava/lang/String;" ) ) );

                while(1)
                {
                    jmethodID get_message_id = jenv->GetMethodID( c, "getMessage", "()Ljava/lang/String;" );
                    if( !get_message_id )  break;

                    xc._message += string_from_jstring( Jstring( Vm::static_vm, (jstring)jenv->CallObjectMethod( x, get_message_id ) ) );
                  //xc._message += replace_regex_ext( string_from_jstring( Jstring( Vm::static_vm, (jstring)jenv->CallObjectMethod( x, get_message_id ) ) ),
                  //                                  "^(.*)\r?\n?$", "\\1" );  // Der Regex greift nicht.
                    

                    jmethodID get_cause_id = jenv->GetMethodID( c, "getCause", "()Ljava/lang/Throwable;" );
                    if( !get_cause_id )  break;

                    jthrowable x2 = (jthrowable)jenv->CallObjectMethod( x, get_cause_id );
                    if( !x2 )  break;

                    xc._message += Z_NL;

                    jenv->DeleteLocalRef( x );
                    x = x2;
                }

                jenv->DeleteLocalRef( c );
            }

            if( x )  jenv->DeleteLocalRef( x );
        }
    }

    //if( xc._message.empty() )  xc._message = "java";

    xc.insert( 1, xc._exception_name );
    xc.insert( 2, xc._message );
    xc.insert( 3, text1 );
    xc.insert( 4, text2 );

    throw xc;
}

//----------------------------------------------------------------------------------Env::throw_java
/*
void Env::throw_java( jclass c, const string& text1 )
{
    string class_name;

    try
    {
        class_name = get_class_name( jenv(), c );
    }
    catch( exception& ) {}

    throw_java( class_name, text1 );
}
*/
//----------------------------------------------------------------------------------Env::add_object
/*
void Env::add_object( Java_idispatch* o )
{ 
    _java_idispatch_list.push_back(NULL); 
    *_java_idispatch_list.rbegin() = o; 

    o->set_global(); 
}
*/
//----------------------------------------------------------------------------------Env::find_class

jclass Env::find_class( const string& name_par )
{
    string name = name_par;

    for( uint i = 0; i < name.length(); i++ )  if( name[i] == '.' )  name[i] = '/';

    if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "java FindClass " << name << '\n' );

    jclass result = _jenv->FindClass( name.c_str() );
    if( !result )  throw_java( "FindClass", name );

    return result;
}

//-------------------------------------------------------------------------------Env::get_method_id

jmethodID Env::get_method_id( jclass cls, const char* name, const char* signature, bool is_static )
{
    jmethodID result;
    
    if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "java get_method_id(\"" << name << "\")\n" );

    if( is_static )
    {
        result = _jenv->GetStaticMethodID( cls, name, signature );
        if( !result )  throw_java( "GetStaticMethodID", get_class_name( jenv(), cls ) + "." + string(name) + signature );
    }
    else
    {
        result = _jenv->GetMethodID( cls, name, signature );
        if( !result )  throw_java( "GetMethodID", get_class_name( jenv(), cls ) + "." + string(name) + signature );
    }

    return result;
}

//----------------------------------------------------------------------------Env::get_object_class

jclass Env::get_object_class( jobject o )
{
    jclass result = _jenv->GetObjectClass( o );
    if( !result )  throw_java( "GetObjectClass" );

    return result;
}

//-------------------------------------------------------------------------Local_frame::Local_frame

Local_frame::Local_frame( JNIEnv* jenv, int capacity ) 
: 
    _jenv(jenv) 
{ 
    int ret = jenv->PushLocalFrame( capacity ); 
    if( ret )  throw_java_ret( ret, "PushLocalFrame" ); 
}

//------------------------------------------------------------------------------------Jvalue::check

void Jvalue::check( char type )
{
    if( _type != type )  throw_xc( "Z-JAVA-107", string( &type, 1 ), string( &_type, 1 ) );
}

//-------------------------------------------------------------------------------------Class::Class

Class::Class( Vm* vm, const char* name )
: 
    Global_jobject(vm)
{ 
    if( name )  load( name ); 
}

//-------------------------------------------------------------------------------------Class::Class

Class::Class( Vm* vm, const string& name )
: 
    Global_jobject(vm)
{ 
    load( name ); 
}

//--------------------------------------------------------------------------------------Class::load

void Class::load( const string& name )
{
    Env e = env();

    jclass jcls = e.find_class( name );
    assign( jcls );
    e->DeleteLocalRef( jcls );

    _name = name;
}

//------------------------------------------------------------------------------Class::new_instance
/*
jobject Class::new_instance()
{
    Env e = env();

}
*/
//--------------------------------------------------------------------------------Class::new_object

jobject Class::new_object( const char* signature, ... )
{
    jobject result = NULL;

    try
    {
        Env e = env();

        va_list args;
        va_start( args, signature );

            if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java NewObjectV(" << signature << ") ...\n" );

            result = e->NewObjectV( get_jclass(), method_id( "<init>", signature ), args );

            if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java NewObjectV(" << signature << ")  OK\n" );
        
        va_end( args );

        if( e->ExceptionCheck() )  e.throw_java( "NewObject", _name );
    }
    catch( Xc& x )  { x.append_text( "Class::new_object " + _name + signature );  throw x; }
    
    return result;
}

//--------------------------------------------------------------------------------Jobject::~Jobject

Jobject::Jobject( Vm* vm, jobject jo )
:
    Has_vm(vm),
    Jvalue( jo )
{
}

//----------------------------------------------------------------------------------Jobject::assign

void Jobject::assign( jobject jo )
{
    _jvalue.l = jo;
}

//------------------------------------------------------------------------------Jobject::get_jclass

jclass Jobject::get_jclass()
{
    if( !_jvalue.l )  throw_xc( "Z-JAVA-106", "get_jclass()" );

    ::jclass result = jenv()->GetObjectClass( _jvalue.l );

    if( !result )  env().throw_java( "GetObjectClass" );

    return result;
}

//-------------------------------------------------------------------------------Jobject::method_id

jmethodID Jobject::method_id( const char* name, const char* signature )
{
    jmethodID result = 0;

    try
    {
        jclass cls = get_jclass();
        if( !cls )  throw_xc( "Z-JAVA-106", string("Methode ") + name );

        result = env().get_method_id( cls, name, signature );
    }
    catch( Xc& x )  { x.append_text( string("Jobject::method_id ") + name + signature );  throw x; }

    return result;
}

//------------------------------------------------------------------------Jobject::static_method_id

jmethodID Jobject::static_method_id( const char* name, const char* signature ) 
{ 
    jmethodID result = 0;

    try
    {
        result = env().get_method_id( get_jclass(), name, signature, true ); 
    }
    catch( Xc& x )  { x.append_text( string("Jobject::static_method_id") + name + signature );  throw x; }

    return result;
}

//------------------------------------------------------------------------Jobject::call_void_method

void Jobject::call_void_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallVoidMethodV(" << name << signature << ") ...\n" );

        e->CallVoidMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallVoidMethodV(" << name << signature << ")  OK\n" );
    
    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( name );
}

//----------------------------------------------------------------------Jobject::call_object_method

jobject Jobject::call_object_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV(" << name << signature << ") ...\n" );

        jobject result = e->CallObjectMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV(" << name << signature << ")  OK\n" );
    
    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( name );

    return result;
}

//----------------------------------------------------------------------Jobject::call_string_method

string Jobject::call_string_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV(" << name << signature << ") ...\n" );

        jstring jstr = (jstring)e->CallObjectMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV(" << name << signature << ")  OK\n" );

    va_end( args );

    string result = string_from_jstring( e, jstr );
    e->DeleteLocalRef( jstr );

    if( e->ExceptionCheck() )  e.throw_java( name );
    
    return result;
}

//----------------------------------------------------------------------Jobject::call_string_method

string Jobject::call_string_method( const jmethodID method_id, ... )
{
    Env     e      = env();
    jobject object = get_jobject();

    va_list args;
    va_start( args, method_id );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV() ...\n" );

        jstring jstr = (jstring)e->CallObjectMethodV( object, method_id, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallObjectMethodV()  OK\n" );

    va_end( args );

    string result = string_from_jstring( e, jstr );
    e->DeleteLocalRef( jstr );

    if( e->ExceptionCheck() )  e.throw_java( "CallObjectMethod" );
    
    return result;
}

//-------------------------------------------------------------------------Jobject::call_int_method

jint Jobject::call_int_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallIntMethodV(" << name << signature << ") ...\n" );

        jint result = e->CallIntMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallIntMethodV(" << name << signature << ")  OK\n" );

    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( name );
    
    return result;
}

//------------------------------------------------------------------------Jobject::call_long_method

jlong Jobject::call_long_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallLongMethodV(" << name << signature << ") ...\n" );

        jlong result = e->CallLongMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallLongMethodV(" << name << signature << ")  OK\n" );

    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( name );
    
    return result;
}

//------------------------------------------------------------------------Jobject::call_bool_method

bool Jobject::call_bool_method( const char* name, const char* signature, ... )
{
    Env       e      = env();
    jobject   object = get_jobject();
    jmethodID method = method_id( name, signature );

    va_list args;
    va_start( args, signature );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallBooleanMethodV(" << name << signature << ") ...\n" );

        jboolean result = e->CallBooleanMethodV( object, method, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallBooleanMethodV(" << name << signature << ")  OK\n" );

    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( name );
    
    return result != 0;
}

//------------------------------------------------------------------------Jobject::call_bool_method

bool Jobject::call_bool_method( jmethodID method_id, ... )
{
    Env     e      = env();
    jobject object = get_jobject();

    va_list args;
    va_start( args, method_id );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallBooleanMethodV() ...\n" );

        jboolean result = e->CallBooleanMethodV( object, method_id, args );

        if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallBooleanMethodV()  OK\n" );

    va_end( args );

    if( e->ExceptionCheck() )  e.throw_java( "CallBooleanMethod" );
    
    return result != 0;
}

//---------------------------------------------------------------Jobject::call_static_object_method

jobject Jobject::call_static_object_method( const char* name, const char* signature, ... )
{
    jobject result = NULL;

    try
    {
        Env       e         = env();
        jclass    cls       = get_jclass();
        jmethodID method_id = static_method_id( name, signature );


        va_list args;
        va_start( args, signature );

            if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallStaticObjectMethod(" << name << signature << ") ...\n" );

            result = e->CallStaticObjectMethodV( cls, method_id, args );

            if( Vm::static_vm && Vm::static_vm->_debug )  Z_LOG( "Java CallStaticObjectMethod(" << name << signature << ")  OK.\n" );

        va_end( args );

        if( e->ExceptionCheck() )  e.throw_java( name );
    }
    catch( Xc& x )  { x.append_text( string("Jobject::call_static_object_method ") + name + signature );  throw x; }

    return result;
}

//----------------------------------------------------------------------Jobject::call_object_method
/*
jobject Jobject::call_object_method( const char* name_and_signature )
{
    Env e = env();
    Local<jclass> cls = get_jclass();

    jmethodID method_id = e.get_method_id( cls, e.get_method_id( cls, name_and_signature );
    
    jobject result = e->CallObjectMethod( cls, method_id );
    if( e->ExceptionOccurred() )  e.throw_java( name_and_signature );

    return result;
}

jobject Jobject::call_object_method( const char* name_and_signature, const Jvalue&, const Jvalue&, const Jvalue& )
{
    Env e = env();
    Local<jclass> cls = get_jclass();

    jmethodID method_id = e.get_method_id( cls, e.get_method_id( cls, name_and_signature );
    
    jobject result = e->CallObjectMethod( cls, method_id, value.get_jvalue() );
    if( e->ExceptionOccurred() )  e.throw_java( name_and_signature );

    return result;
}
*/
//---------------------------------------------------------------------------Global_jobject::assign

void Global_jobject::assign( jobject jo )
{
    if( !_jvalue.l && !jo )  return;

    Env e = env();

    if( _jvalue.l )
    {
        e->DeleteGlobalRef( _jvalue.l );
        _jvalue.l = NULL;
    }

    if( jo )
    {
        _jvalue.l = e->NewGlobalRef( jo );
        if( !_jvalue.l )  e.throw_java( "NewGlobalRef" );
    }
}

//--------------------------------------------------------------------Local_jobject::~Local_jobject

Local_jobject::~Local_jobject()
{ 
    if( _call_close_at_end  &&  get_jobject() )
    {
        try
        {
            call_void_method( "close", "()V" );
        }
        catch( exception& x )
        {
            Z_LOG( __FUNCTION__ << ".close() ==> ERROR " << x.what() << '\n' );
        }
    }

    assign( NULL ); 
}

//----------------------------------------------------------------------------Local_jobject::assign

void Local_jobject::assign( jobject jo )
{
    if( _jvalue.l )
    {
        jenv()->DeleteLocalRef( _jvalue.l );
        _jvalue.l = NULL;
    }

    _jvalue.l = jo;
}

//--------------------------------------------------------------------------------Jstring::~Jstring

Jstring::~Jstring()
{ 
    assign( (jstring)NULL );
}

//----------------------------------------------------------------------------------Jstring::assign

void Jstring::assign( jstring jstr )
{
    if( _jvalue.l )  jenv()->DeleteLocalRef( _jvalue.l ); 
    
    _jvalue.l = jstr;
}

//----------------------------------------------------------------------------------Jstring::assign

void Jstring::assign( const char* str, size_t length )
{
    assign( jstring_from_string( jenv(), str, length ) );
}

//------------------------------------------------------Read_jstring_critical::Read_jstring_critical

Read_jstring_critical::Read_jstring_critical( JNIEnv* jenv, jstring jstr )
:
    _jenv(jenv),
    _jstring(jstr)
{
    jboolean is_copy;

    _jchars = jenv->GetStringCritical( _jstring, &is_copy );
    if( !_jchars )  Env(_jenv).throw_java( "GetStringCritical()" );
}

//----------------------------------------------------Read_jstring_critical::~Read_jstring_critical

Read_jstring_critical::~Read_jstring_critical()
{
    _jenv->ReleaseStringCritical( _jstring, _jchars );
}

//-----------------------------------------------------------------------------------Method::assign

void Method::assign( const char* name, const char* signature, bool is_static )
{
    _id = NULL;

    if( name[0] == '\0'  &&  signature[0] == '\0' )  return;

    Z_DEBUG_ONLY( _name_and_signature = string(name) + signature; )

    _id = is_static? _class->static_method_id( name, signature ) : _class->method_id( name, signature );
    //_id = env().get_method_id( _jclass, name, signature );
}

//-----------------------------------------------------------------------------------Method::assign

void Method::assign( const string& name_and_signature, bool is_static )
{
    _id = NULL;

    if( name_and_signature.empty() )  return;

    size_t pos = name_and_signature.find( '(' );
    if( pos == string::npos )  pos = name_and_signature.length();
    
    assign( name_and_signature.substr(0,pos).c_str(), name_and_signature.c_str()+pos, is_static );
}

//---------------------------------------------------------------------------------------Call::Call
/*
Call::Call( Env& e, jobject jo, const char* name_and_signature )
:
    _zero_(this+1)
{
    if( !jo )  throw_xc( "Z-JAVA-106" );

    _jclass = e->GetObjectClass( jo );
    if( !_jclass )  e.throw_java( "GetObjectClass" );

    _method_id = e.get_method_id( _jclass, name_and_signature );
}
*/
//---------------------------------------------------------------------------------------Call::Call
/*
Call::Call( jclass cls, const char* name_and_signature )
:
    _zero_(this+1)
{
    _is_static = true;

    if( !cls )  throw_xc( "Z-JAVA-106" );

    _method_id = e.get_method_id( _jclass, name_and_signature, true );
}
*/
//---------------------------------------------------------------------Call::call_unchecked_jobject
/*
jobject Call::call_unchecked_jobject()
{
    jobject result = env()->CallObjectMethodA( _jobject, _method_id, _jvalues );
    //if( env()->ExceptionCheck() )  env().throw_java( _name_and_signature );
}
*/
//-------------------------------------------------------------------------------------------------


} //namespace java
} //namespace zschimmer
