// zschimmer.h                                      2000 Joacim Zschimmer
// $Id: base.h 4169 2006-03-08 10:19:39Z jz $

#ifndef __ZSCHIMMER_MESSAGE_H
#define __ZSCHIMMER_MESSAGE_H

namespace zschimmer {

//------------------------------------------------------------------------------insert_into_message

struct Message_string;
struct Object;

void        insert_into_message( Message_string* m, int index, int64    ) throw();
inline void insert_into_message( Message_string* m, int index, long   v ) throw()                   { return insert_into_message( m, index, (int64)v ); }
inline void insert_into_message( Message_string* m, int index, ulong  v ) throw()                   { return insert_into_message( m, index, (int64)v ); }
inline void insert_into_message( Message_string* m, int index, int    v ) throw()                   { return insert_into_message( m, index, (int64)v ); }
inline void insert_into_message( Message_string* m, int index, uint   v ) throw()                   { return insert_into_message( m, index, (int64)v ); }
inline void insert_into_message( Message_string* m, int index, short  v ) throw()                   { return insert_into_message( m, index, (int64)v ); }
inline void insert_into_message( Message_string* m, int index, unsigned short v ) throw()           { return insert_into_message( m, index, (int64)v ); }
//inline void insert_into_message( Message_string* m, int index, size_t v ) throw()                   { return insert_into_message( m, index, (int64)v ); }

void        insert_into_message( Message_string* m, int index, double               ) throw();
void        insert_into_message( Message_string* m, int index, const exception&     ) throw();
void        insert_into_message( Message_string* m, int index, const Object&        ) throw();
void        insert_into_message( Message_string* m, int index, const String_stream& ) throw();

//-----------------------------------------------------------------------------------Message_string

struct Message_string
{
    Message_string( const string& code );

    Message_string( const char* code );

    template< typename T1 >
    Message_string( const string& code, const T1& p1 )
    {
        set_code( code );
        insert( 1, p1 );
    }
    
    template< typename T1, typename T2 >
    Message_string( const string& code, const T1& p1, const T2& p2 )
    {
        set_code( code );
        insert( 1, p1 );
        insert( 2, p2 );
    }

    template< typename T1, typename T2, typename T3 >
    Message_string( const string& code, const T1& p1, const T2& p2, const T3& p3 )
    {
        set_code( code );
        insert( 1, p1 );
        insert( 3, p2 );
        insert( 2, p3 );
    }

    void                        set_code                    ( const string& );
    void                        set_code                    ( const char* );

    template< typename T >
    void                        insert                      ( int index, const T& value ) throw()   { insert_into_message( this, index, value ); }

    void                        insert                      ( int index, const string& value ) throw()  { insert_string( index, value ); }
    void                        insert                      ( int index, const char*   value ) throw()  { insert_string( index, value ); }

    void                        insert_string               ( int index, const string& ) throw();
    void                        insert_string               ( int index, const char* ) throw();

                                operator const string&      () const                                { return _string; }
    const string&               as_string                   () const                                { return _string; }

    Z_NORETURN void             throw_xc                    ( const char* code ) const;
    Z_NORETURN void             throw_xc                    ( const string& code ) const;

  private:
    string                     _string;
};

//-----------------------------------------------------------------------------------message_string

template< typename CODE_TYPE >
string message_string( const CODE_TYPE& code ) 
{
    Message_string result ( code );
    return result.as_string();
}


template< typename CODE_TYPE, typename T1 >
string message_string( const CODE_TYPE& code, const T1& p1 ) 
{
    Message_string result ( code );

    result.insert( 1, p1 ); 

    return result.as_string();
}


template< typename CODE_TYPE, typename T1, typename T2 >
string message_string( const CODE_TYPE& code, const T1& p1, const T2& p2 ) 
{
    Message_string result ( code );

    result.insert( 1, p1 ); 
    result.insert( 2, p2 ); 

    return result.as_string();
}


template< typename CODE_TYPE, typename T1, typename T2, typename T3 >
string message_string( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3 ) 
{
    Message_string result ( code );

    result.insert( 1, p1 ); 
    result.insert( 2, p2 ); 
    result.insert( 3, p3 ); 

    return result.as_string();
}


template< typename CODE_TYPE, typename T1, typename T2, typename T3, typename T4 >
string message_string( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3, const T4& p4 ) 
{
    Message_string result ( code );

    result.insert( 1, p1 ); 
    result.insert( 2, p2 ); 
    result.insert( 3, p3 ); 
    result.insert( 3, p4 ); 

    return result.as_string();
}

//-----------------------------------------------------------------------------------------throw_xc

template< typename CODE_TYPE >
Z_NORETURN void throw_xc( const CODE_TYPE& code ) 
{
    Message_string message ( code );
    message.throw_xc( code );
}


template< typename CODE_TYPE, typename T1 >
Z_NORETURN void throw_xc( const CODE_TYPE& code, const T1& p1 ) 
{
    Message_string message ( code );

    message.insert( 1, p1 ); 

    message.throw_xc( code );
}


template< typename CODE_TYPE, typename T1, typename T2 >
Z_NORETURN void throw_xc( const CODE_TYPE& code, const T1& p1, const T2& p2 ) 
{
    Message_string message ( code );

    message.insert( 1, p1 ); 
    message.insert( 2, p2 ); 

    message.throw_xc( code );
}


template< typename CODE_TYPE, typename T1, typename T2, typename T3 >
Z_NORETURN void throw_xc( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3 ) 
{
    Message_string message ( code );

    message.insert( 1, p1 ); 
    message.insert( 2, p2 ); 
    message.insert( 3, p3 ); 

    message.throw_xc( code );
}


template< typename CODE_TYPE, typename T1, typename T2, typename T3, typename T4 >
Z_NORETURN void throw_xc( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3, const T4& p4 ) 
{
    Message_string message ( code );

    message.insert( 1, p1 ); 
    message.insert( 2, p2 ); 
    message.insert( 3, p3 ); 
    message.insert( 3, p4 ); 

    message.throw_xc( code );
}

//---------------------------------------------------------------------------------------Source_pos

struct Source_pos
{
    explicit               Source_pos                  ( const string& source = "", int line = 0, int col = 0 ) : _source(source), _line(line), _col(col) {}

    string                 _source;
    int                    _line;
    int                    _col;
};

//-----------------------------------------------------------------------------------------------Xc

struct Xc : std::exception
{
                            Xc                          ();
                            Xc                          ( const exception& );
                            Xc                          ( const string& code );
                            Xc                          ( const char* code );
                            Xc                          ( const char* code, const Message_string& );
                            Xc                          ( const string& code, const Message_string& );
                            Xc                          ( const _com_error& );                      // Implementiert in z_com.cxx
                            //Xc                          ( const char* code, const char* ins1 = NULL, const char* ins2 = NULL, const char* ins3 = NULL, const char* ins4 = NULL );


    template< typename CODE_TYPE, typename T1 >
    Xc( const CODE_TYPE& code, const T1& p1 ) 
    : 
        _code( code ), 
        _return_code(0)
    {
        Message_string message ( code );

        message.insert( 1, p1 ); 

        set_what( message.as_string() );
    }


    template< typename CODE_TYPE, typename T1, typename T2 >
    Xc( const CODE_TYPE& code, const T1& p1, const T2& p2 ) 
    : 
        _code( code ), 
        _return_code(0)
    {
        Message_string message ( code );

        message.insert( 1, p1 ); 
        message.insert( 2, p2 ); 

        set_what( message.as_string() );
    }


    template< typename CODE_TYPE, typename T1, typename T2, typename T3 >
    Xc( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3 ) 
    : 
        _code( code ), 
        _return_code(0)
    {
        Message_string message ( code );

        message.insert( 1, p1 ); 
        message.insert( 2, p2 ); 
        message.insert( 3, p3 ); 

        set_what( message.as_string() );
    }


    template< typename CODE_TYPE, typename T1, typename T2, typename T3, typename T4 >
    Xc( const CODE_TYPE& code, const T1& p1, const T2& p2, const T3& p3, const T4& p4 ) 
    : 
        _code( code ), 
        _return_code(0)
    {
        Message_string message ( code );

        message.insert( 1, p1 ); 
        message.insert( 2, p2 ); 
        message.insert( 3, p3 ); 
        message.insert( 3, p4 ); 

        set_what( message.as_string() );
    }

                            Xc                          ( const char* code, const string& text, const Source_pos& );
                           ~Xc                          () throw();

    const char*             what                        () const throw();
    void                set_what                        ( const string& );

    string                  name                        () const                                    { return _name; }
    void                set_name                        ( const string& name )                      { _name = name; }
    
    string                  code                        () const                                    { return _code; }
    void                set_code                        ( const char* );
    void                set_code                        ( const string& code )                      { set_code( code.c_str() ); }
    
    void                    insert                      ( int nr, const char* );
    void                    insert                      ( int nr, const string& s )                 { insert( nr, s.c_str() ); }
    void                    insert                      ( int nr, int );
    void                    append_text                 ( const string& text )                      { append_text( text.c_str() ); }
    void                    append_text                 ( const char* );

    friend ostream&         operator <<                 ( ostream&, const Xc& );


    string                 _name;
    string                 _code;
    string                 _what;
    int                    _return_code;                // Returncode bei Programmende (0: unbestimmt)

  private:
    void                    set_what_by_code            ( const char* code );
};

//----------------------------------------------------------------------------------------Eof_error

struct Eof_error : Xc
{
                            Eof_error                   ( const char* code, const char* text = "" ) : Xc( code, text ) {}
};

/*
Z_NORETURN void             throw_xc                    ( const char* error_code, const char* = NULL, const char* = NULL, const char* = NULL );
Z_NORETURN void             throw_xc                    ( const char* error_code, const string& );
inline Z_NORETURN void      throw_xc                    ( const char* error_code, const string& a, const string& b )    { throw_xc( error_code, a.c_str(), b.c_str() ); }
inline Z_NORETURN void      throw_xc                    ( const char* error_code, const string& a, const string& b, const string& c )    { throw_xc( error_code, a.c_str(), b.c_str(), c.c_str() ); }
Z_NORETURN void             throw_xc                    ( const char* error_code, int64 );
Z_NORETURN void             throw_xc                    ( const char* error_code, int64, int64 );
*/
Z_NORETURN void             throw_xc                    ( const char* error_code, const string&, const Source_pos& );
Z_NORETURN void             throw_xc                    ( const char* error_code, const char*  , const Source_pos& );
inline Z_NORETURN void      throw_xc                    ( const Xc& x )                                                 { throw x; }
Z_NORETURN void             throw_eof                   ();
Z_NORETURN void             throw_errno                 ( int errn, const char* = NULL, const char* = NULL, const char* = NULL );
Z_NORETURN void             throw_socket                ( int errn, const char* = NULL, const char* = NULL, const char* = NULL );
Z_NORETURN void             throw_pattern               ( const char* pattern, int error, const char* function_name, const char* text1 = NULL, const char* text2 = NULL );
Z_NORETURN void             throw_mswin                 ( int mswin_error, const char* function = "", const char* = NULL, const char* = NULL );
inline Z_NORETURN void      throw_mswin                 ( int mswin_error, const char* function, const string& a ) { throw_mswin( mswin_error, function, a.c_str() ); }
Z_NORETURN void             throw_mswin                 ( const char* function = "", const char* ins1 = NULL, const char* ins2 = NULL );
inline Z_NORETURN void      throw_mswin                 ( const char* function, const string& ins1 )                       { throw_mswin( function, ins1.c_str() ); }
Z_NORETURN inline void      throw_overflow              ( const char* error_code, const char* a = "", const char* b = NULL )  { throw_xc( error_code, a, b ); }

const char*                 get_error_text              ( const char* code );
inline const char*          get_error_text              ( const string& code )                      { return get_error_text( code.c_str() ); }

string                      get_mswin_msg_text          ( int code );

string                      errno_code                  ( int errn = errno );

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer

#endif
