// $Id: z_process.h 4231 2006-04-03 18:20:24Z jz $

#ifndef __ZSCHIMMER_PROCESS_H
#define __ZSCHIMMER_PROCESS_H

#include "z_com_server.h"


namespace zschimmer
{

bool                            try_kill_process_immediately( int pid, bool kill_childs = false );
void                            kill_process_immediately    ( int pid, bool kill_childs = false );

//-------------------------------------------------------------------------------------Process_base
    
struct Process_base : Object
{
    typedef stdext::hash_map<string,string> Environment;


    virtual                    ~Process_base                ();

    virtual int                 pid                         ()                                      { return _pid; }
    virtual string              command_line                () const                                { return _command_line; }
    string                      obj_name                    () const                                { return S() << "Process(pid=" << _pid << ",cmd=" << _command_line << ")"; }

    bool                        started                     () const                                { return _pid != 0; }
    void                        assert_not_started          ()                                      { if( started()    )  throw_xc( "Z-PROCESS-002" ); }
    void                        assert_terminated           ()                                      { if( !_terminated )  throw_xc( "Z-PROCESS-001", _pid ); }
    virtual bool            set_priority                    ( int )                                 = 0;
    virtual int                 priority                    ()                                      = 0;
    virtual bool            set_priority_class              ( const string& )                       = 0;
    virtual string              priority_class              ()                                      = 0;
    virtual bool                terminated                  ()                                      = 0;
    virtual int                 exit_code                   ()                                      = 0;
    virtual int                 termination_signal          ()                                      { assert_terminated(); return 0; }
    virtual void                kill                        ( int )                                 = 0;
  //virtual void            set_with_console_window         ( bool )                                {}

    void                    set_stdout_path                 ( const string& path )                  { assert_not_started();  _stdout_path = path; }
    string                      stdout_path                 ()                                      { return _stdout_path; }
    void                    set_stderr_path                 ( const string& path )                  { assert_not_started();  _stderr_path = path; }
    string                      stderr_path                 ()                                      { return _stderr_path; }
    void                    set_environment_entry           ( const string& name, const string& value ) { _environment[ name ] = value; }
  //virtual Environment&        environment                 ()                                      { return _environment; }


  //STDMETHODIMP                Raise_priority              ( int diff, VARIANT_BOOL* result )      { Z_COM_IMPLEMENT( *result = _process.raise_priority( diff )? VARIANT_TRUE : VARIANT_FALSE ); }
  //STDMETHODIMP                Lower_priority              ( int diff, VARIANT_BOOL* result )      { return Raise_priority( -diff, result ); }

    virtual STDMETHODIMP    get_Pid                         ( int* result )                         { Z_COM_IMPLEMENT( *result = pid() ) }
    virtual STDMETHODIMP    get_Terminated                  ( VARIANT_BOOL* result )                { Z_COM_IMPLEMENT( *result = terminated()? VARIANT_TRUE: VARIANT_FALSE ) }
    virtual STDMETHODIMP    get_Exit_code                   ( int* result )                         { Z_COM_IMPLEMENT( *result = exit_code() ) }
    virtual STDMETHODIMP    get_Termination_signal          ( int* result )                         { Z_COM_IMPLEMENT( *result = termination_signal() ) }
    virtual STDMETHODIMP    put_Priority                    ( int priority )                        { Z_COM_IMPLEMENT( set_priority( priority ) ) } 
    virtual STDMETHODIMP    get_Priority                    ( int* result )                         { Z_COM_IMPLEMENT( *result = priority() ) } 
    virtual STDMETHODIMP    put_Priority_class              ( BSTR priority )                       { Z_COM_IMPLEMENT( set_priority_class( com::string_from_bstr( priority ) ) ) } 
    virtual STDMETHODIMP    get_Priority_class              ( BSTR* result )                        { Z_COM_IMPLEMENT( return com::String_to_bstr( priority_class(), result ) ) } 
    virtual STDMETHODIMP        Kill                        ( int kill_signal )                     { Z_COM_IMPLEMENT( kill( kill_signal ) ); }

  protected:
    int                        _pid;
    bool                       _terminated;
    string                     _command_line;               // Frs Log
    string                     _stdout_path;
    string                     _stderr_path;
    Environment                _environment;
};

//-------------------------------------------------------------------------------------------------

} // namespace zschimmer




#ifdef Z_WINDOWS

#   include "z_windows_process.h"

    namespace zschimmer
    {
        typedef windows::Process Process;
    }

#else

#   include "z_posix_process.h"

    namespace zschimmer
    {
        typedef posix::Process Process;
    }

#endif

//-------------------------------------------------------------------------------------------------

#endif
